from test_autogalaxy.simulators.imaging import simulators

"""
Welcome to the PyAutoGalaxy dataset generator. Here, we'll make the datasets used to test and profile PyAutoGalaxy.

This consists of the following sets of images:

- An image where the galaxy is an elliptical Dev Vaucouleurs profile.
- An image where the galaxy is a bulge (Dev Vaucouleurs) + Envelope (Exponential) profile.
- An image where there are two galaxies both composed of Sersic bulges.
- An image where the galaxy is an elliptical Dev Vaucouleurs profile with a centre offset from the image centre.

Each image is generated for up to 5 instrument, VRO, Euclid, HST, HST, Keck AO.
"""

instruments = ["vro", "euclid", "hst"]  # , "hst_up", "ao"]

# To simulate each galaxy, we pass it an instrument and call its simulate function.

for instrument in instruments:

    simulators.simulate__galaxy_x1__dev_vaucouleurs(instrument=instrument)
    simulators.simulate__galaxy_x1__bulge_disk(instrument=instrument)
    simulators.simulate__galaxy_x2__sersics(instrument=instrument)
    simulators.simulate__galaxy_x1__dev_vaucouleurs__offset_centre(
        instrument=instrument
    )



import autofit as af


class PipelineDataset(af.Pipeline):
    def run(self, dataset, mask, info=None):
        def runner(phase, results):
            return phase.run(dataset=dataset, results=results, mask=mask, info=info)

        return self.run_function(runner)



from autoarray.plot import plotters
from autogalaxy.plot import lensing_plotters


@lensing_plotters.set_include_and_sub_plotter
@plotters.set_subplot_filename
def subplot_inversion(
    inversion,
    image_positions=None,
    source_positions=None,
    grid=None,
    light_profile_centres=None,
    mass_profile_centres=None,
    critical_curves=None,
    caustics=None,
    image_pixel_indexes=None,
    source_pixel_indexes=None,
    include=None,
    sub_plotter=None,
):

    number_subplots = 6

    ratio = float(
        (
            inversion.mapper.grid.scaled_maxima[1]
            - inversion.mapper.grid.scaled_minima[1]
        )
        / (
            inversion.mapper.grid.scaled_maxima[0]
            - inversion.mapper.grid.scaled_minima[0]
        )
    )

    if sub_plotter.figure.aspect in "square":
        aspect_inv = ratio
    elif sub_plotter.figure.aspect in "auto":
        aspect_inv = 1.0 / ratio
    elif sub_plotter.figure.aspect in "equal":
        aspect_inv = 1.0

    sub_plotter.open_subplot_figure(number_subplots=number_subplots)

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=1)

    reconstructed_image(
        inversion=inversion,
        image_positions=image_positions,
        grid=grid,
        light_profile_centres=light_profile_centres,
        mass_profile_centres=mass_profile_centres,
        critical_curves=critical_curves,
        include=include,
        plotter=sub_plotter,
    )

    sub_plotter.setup_subplot(
        number_subplots=number_subplots, subplot_index=2, aspect=aspect_inv
    )

    reconstruction(
        inversion=inversion,
        source_positions=source_positions,
        caustics=caustics,
        image_pixel_indexes=image_pixel_indexes,
        source_pixel_indexes=source_pixel_indexes,
        include=include,
        plotter=sub_plotter,
    )

    sub_plotter.setup_subplot(
        number_subplots=number_subplots, subplot_index=3, aspect=aspect_inv
    )

    errors(
        inversion=inversion,
        image_pixel_indexes=image_pixel_indexes,
        source_pixel_indexes=source_pixel_indexes,
        caustics=caustics,
        include=include,
        plotter=sub_plotter,
    )

    sub_plotter.setup_subplot(
        number_subplots=number_subplots, subplot_index=4, aspect=aspect_inv
    )

    residual_map(
        inversion=inversion,
        image_pixel_indexes=image_pixel_indexes,
        source_pixel_indexes=source_pixel_indexes,
        caustics=caustics,
        include=include,
        plotter=sub_plotter,
    )

    sub_plotter.setup_subplot(
        number_subplots=number_subplots, subplot_index=5, aspect=aspect_inv
    )

    chi_squared_map(
        inversion=inversion,
        image_pixel_indexes=image_pixel_indexes,
        source_pixel_indexes=source_pixel_indexes,
        caustics=caustics,
        include=include,
        plotter=sub_plotter,
    )

    sub_plotter.setup_subplot(
        number_subplots=number_subplots, subplot_index=6, aspect=aspect_inv
    )

    regularization_weights(
        inversion=inversion,
        source_positions=source_positions,
        image_pixel_indexes=image_pixel_indexes,
        source_pixel_indexes=source_pixel_indexes,
        caustics=caustics,
        include=include,
        plotter=sub_plotter,
    )

    sub_plotter.output.subplot_to_figure()

    sub_plotter.figure.close()


def individuals(
    inversion,
    image_positions=None,
    source_positions=None,
    grid=None,
    light_profile_centres=None,
    mass_profile_centres=None,
    critical_curves=None,
    caustics=None,
    plot_reconstructed_image=False,
    plot_reconstruction=False,
    plot_errors=False,
    plot_residual_map=False,
    plot_normalized_residual_map=False,
    plot_chi_squared_map=False,
    plot_regularization_weight_map=False,
    plot_interpolated_reconstruction=False,
    plot_interpolated_errors=False,
    include=None,
    plotter=None,
):
    """Plot the model datas_ of an analysis, using the *Fitter* class object.

    The visualization and output type can be fully customized.

    Parameters
    -----------
    fit : autogalaxy.lens.fitting.Fitter
        Class containing fit between the model datas_ and observed lens datas_ (including residual_map, chi_squared_map etc.)
    output_path : str
        The path where the datas_ is output if the output_type is a file format (e.g. png, fits)
    output_format : str
        How the datas_ is output. File formats (e.g. png, fits) output the datas_ to harddisk. 'show' displays the datas_ \
        in the python interpreter window.
    """

    if plot_reconstructed_image:

        reconstructed_image(
            inversion=inversion,
            image_positions=image_positions,
            grid=grid,
            light_profile_centres=light_profile_centres,
            mass_profile_centres=mass_profile_centres,
            critical_curves=critical_curves,
            include=include,
            plotter=plotter,
        )

    if plot_reconstruction:

        reconstruction(
            inversion=inversion,
            source_positions=source_positions,
            caustics=caustics,
            include=include,
            plotter=plotter,
        )

    if plot_errors:

        errors(inversion=inversion, include=include, caustics=caustics, plotter=plotter)

    if plot_residual_map:

        residual_map(
            inversion=inversion, include=include, caustics=caustics, plotter=plotter
        )

    if plot_normalized_residual_map:

        normalized_residual_map(
            inversion=inversion, caustics=caustics, include=include, plotter=plotter
        )

    if plot_chi_squared_map:

        chi_squared_map(
            inversion=inversion, caustics=caustics, include=include, plotter=plotter
        )

    if plot_regularization_weight_map:

        regularization_weights(
            inversion=inversion, caustics=caustics, include=include, plotter=plotter
        )

    if plot_interpolated_reconstruction:

        interpolated_reconstruction(
            inversion=inversion,
            source_positions=source_positions,
            caustics=caustics,
            include=include,
            plotter=plotter,
        )

    if plot_interpolated_errors:

        interpolated_errors(
            inversion=inversion, caustics=caustics, include=include, plotter=plotter
        )


@lensing_plotters.set_include_and_plotter
@plotters.set_labels
def reconstructed_image(
    inversion,
    image_positions=None,
    grid=None,
    light_profile_centres=None,
    mass_profile_centres=None,
    critical_curves=None,
    include=None,
    plotter=None,
):

    plotter.plot_array(
        array=inversion.mapped_reconstructed_image,
        mask=include.mask_from_grid(grid=inversion.mapper.grid),
        positions=image_positions,
        grid=grid,
        light_profile_centres=light_profile_centres,
        mass_profile_centres=mass_profile_centres,
        critical_curves=critical_curves,
        include_origin=include.origin,
    )


@lensing_plotters.set_include_and_plotter
@plotters.set_labels
def reconstruction(
    inversion,
    source_positions=None,
    caustics=None,
    image_pixel_indexes=None,
    source_pixel_indexes=None,
    include=None,
    plotter=None,
):

    source_pixel_values = inversion.mapper.reconstructed_pixelization_from_solution_vector(
        solution_vector=inversion.reconstruction
    )

    plotter.plot_mapper(
        mapper=inversion.mapper,
        source_pixel_values=source_pixel_values,
        positions=source_positions,
        caustics=caustics,
        image_pixel_indexes=image_pixel_indexes,
        source_pixel_indexes=source_pixel_indexes,
        include_origin=include.origin,
        include_pixelization_grid=include.inversion_pixelization_grid,
        include_grid=include.inversion_grid,
        include_border=include.inversion_border,
    )


@lensing_plotters.set_include_and_plotter
@plotters.set_labels
def errors(
    inversion,
    source_positions=None,
    caustics=None,
    image_pixel_indexes=None,
    source_pixel_indexes=None,
    include=None,
    plotter=None,
):

    source_pixel_values = inversion.mapper.reconstructed_pixelization_from_solution_vector(
        solution_vector=inversion.errors
    )

    plotter.plot_mapper(
        mapper=inversion.mapper,
        source_pixel_values=source_pixel_values,
        positions=source_positions,
        caustics=caustics,
        image_pixel_indexes=image_pixel_indexes,
        source_pixel_indexes=source_pixel_indexes,
        include_origin=include.origin,
        include_pixelization_grid=include.inversion_pixelization_grid,
        include_grid=include.inversion_grid,
        include_border=include.inversion_border,
    )


@lensing_plotters.set_include_and_plotter
@plotters.set_labels
def residual_map(
    inversion,
    source_positions=None,
    caustics=None,
    image_pixel_indexes=None,
    source_pixel_indexes=None,
    include=None,
    plotter=None,
):

    source_pixel_values = inversion.mapper.reconstructed_pixelization_from_solution_vector(
        solution_vector=inversion.residual_map
    )

    plotter.plot_mapper(
        mapper=inversion.mapper,
        source_pixel_values=source_pixel_values,
        positions=source_positions,
        caustics=caustics,
        image_pixel_indexes=image_pixel_indexes,
        source_pixel_indexes=source_pixel_indexes,
        include_origin=include.origin,
        include_pixelization_grid=include.inversion_pixelization_grid,
        include_grid=include.inversion_grid,
        include_border=include.inversion_border,
    )


@lensing_plotters.set_include_and_plotter
@plotters.set_labels
def normalized_residual_map(
    inversion,
    source_positions=None,
    caustics=None,
    image_pixel_indexes=None,
    source_pixel_indexes=None,
    include=None,
    plotter=None,
):

    source_pixel_values = inversion.mapper.reconstructed_pixelization_from_solution_vector(
        solution_vector=inversion.normalized_residual_map
    )

    plotter.plot_mapper(
        mapper=inversion.mapper,
        source_pixel_values=source_pixel_values,
        positions=source_positions,
        caustics=caustics,
        image_pixel_indexes=image_pixel_indexes,
        source_pixel_indexes=source_pixel_indexes,
        include_origin=include.origin,
        include_pixelization_grid=include.inversion_pixelization_grid,
        include_grid=include.inversion_grid,
        include_border=include.inversion_border,
    )


@lensing_plotters.set_include_and_plotter
@plotters.set_labels
def chi_squared_map(
    inversion,
    source_positions=None,
    caustics=None,
    image_pixel_indexes=None,
    source_pixel_indexes=None,
    include=None,
    plotter=None,
):

    source_pixel_values = inversion.mapper.reconstructed_pixelization_from_solution_vector(
        solution_vector=inversion.chi_squared_map
    )

    plotter.plot_mapper(
        mapper=inversion.mapper,
        source_pixel_values=source_pixel_values,
        positions=source_positions,
        caustics=caustics,
        image_pixel_indexes=image_pixel_indexes,
        source_pixel_indexes=source_pixel_indexes,
        include_origin=include.origin,
        include_pixelization_grid=include.inversion_pixelization_grid,
        include_grid=include.inversion_grid,
        include_border=include.inversion_border,
    )


@lensing_plotters.set_include_and_plotter
@plotters.set_labels
def regularization_weights(
    inversion,
    source_positions=None,
    caustics=None,
    image_pixel_indexes=None,
    source_pixel_indexes=None,
    include=None,
    plotter=None,
):

    regularization_weights = inversion.regularization.regularization_weights_from_mapper(
        mapper=inversion.mapper
    )

    plotter.plot_mapper(
        mapper=inversion.mapper,
        source_pixel_values=regularization_weights,
        positions=source_positions,
        caustics=caustics,
        image_pixel_indexes=image_pixel_indexes,
        source_pixel_indexes=source_pixel_indexes,
        include_origin=include.origin,
        include_pixelization_grid=include.inversion_pixelization_grid,
        include_grid=include.inversion_grid,
        include_border=include.inversion_border,
    )


@lensing_plotters.set_include_and_plotter
@plotters.set_labels
def interpolated_reconstruction(
    inversion,
    source_positions=None,
    grid=None,
    caustics=None,
    include=None,
    plotter=None,
):

    plotter.plot_array(
        array=inversion.interpolated_reconstruction_from_shape_2d(),
        positions=source_positions,
        grid=grid,
        caustics=caustics,
        include_origin=include.origin,
    )


@lensing_plotters.set_include_and_plotter
@plotters.set_labels
def interpolated_errors(
    inversion,
    source_positions=None,
    grid=None,
    caustics=None,
    include=None,
    plotter=None,
):

    plotter.plot_array(
        array=inversion.interpolated_errors_from_shape_2d(),
        positions=source_positions,
        grid=grid,
        critical_curves=caustics,
        include_origin=include.origin,
        include_border=include.border,
    )



import inspect

from autofit.mapper.prior_model.prior_model import PriorModel
from autogalaxy.galaxy import Galaxy
from autogalaxy.profiles import light_profiles
from autogalaxy.profiles import mass_profiles


def is_light_profile_class(cls):
    """
    Parameters
    ----------
    cls
        Some object

    Returns
    -------
    bool: is_light_profile_class
        True if cls is a class that inherits from light profile
    """
    return inspect.isclass(cls) and issubclass(cls, light_profiles.LightProfile)


def is_mass_profile_class(cls):
    """
    Parameters
    ----------
    cls
        Some object

    Returns
    -------
    bool: is_mass_profile_class
        True if cls is a class that inherits from mass profile
    """
    return inspect.isclass(cls) and issubclass(cls, mass_profiles.MassProfile)


class GalaxyModel(PriorModel):
    """
    @DynamicAttrs
    """

    def __init__(
        self,
        redshift,
        pixelization=None,
        regularization=None,
        hyper_galaxy=None,
        **kwargs
    ):
        """Class to produce Galaxy instances from sets of profile classes and other model-fitting attributes (e.g. \
         pixelizations, regularization schemes, hyper_galaxies-galaxyes) using the model mapper.

        Parameters
        ----------
        light_profile_classes: [LightProfile]
            The _LightProfile_ classes for which model light profile instances are generated for this galaxy model.
        mass_profile_classes: [MassProfile]
            The _MassProfile_ classes for which model mass profile instances are generated for this galaxy model.
        redshift : float | Type[g.Redshift]
            The redshift of this model galaxy.
        model_redshift : bool
            If *True*, the galaxy redshift will be treated as a free-parameter that is fitted for by the non-linear \
            search.
        pixelization : Pixelization
            The pixelization used to reconstruct the galaxy light and fit the observed if using an inversion.
        regularization : Regularization
            The regularization-scheme used to regularization reconstruct the galaxy light when fitting the observed \
            if using an inversion.
        hyper_galaxy : HyperGalaxy
            A model hyper_galaxies-galaxy used for scaling the observed grid's noise_map.
        """

        super().__init__(
            Galaxy,
            redshift=redshift,
            pixelization=pixelization,
            regularization=regularization,
            hyper_galaxy=hyper_galaxy,
            **kwargs
        )
        profile_models = []

        for name, prior_model in self.prior_model_tuples:
            cls = prior_model.cls
            if is_mass_profile_class(cls) or is_light_profile_class(cls):
                profile_models.append(prior_model)

        if pixelization is not None and regularization is None:
            raise AssertionError(
                "If the galaxy prior has a pixelization, it must also have a "
                "regularization."
            )
        if pixelization is None and regularization is not None:
            raise AssertionError(
                "If the galaxy prior has a regularization, it must also have a "
                "pixelization."
            )



import autofit as af
import autogalaxy as ag
from test_autogalaxy.integration.tests.imaging import runner

test_type = "features"
test_name = "grid_search"
data_label = "galaxy_x1__dev_vaucouleurs"
instrument = "vro"


def make_pipeline(name, folders, search=af.DynestyStatic()):

    bulge = af.PriorModel(ag.lp.EllipticalDevVaucouleurs)

    bulge.centre_0 = af.UniformPrior(lower_limit=-0.01, upper_limit=0.01)
    bulge.centre_1 = af.UniformPrior(lower_limit=-0.01, upper_limit=0.01)
    bulge.axis_ratio = af.UniformPrior(lower_limit=0.79, upper_limit=0.81)
    bulge.phi = af.UniformPrior(lower_limit=-1.0, upper_limit=1.0)
    bulge.intensity = af.UniformPrior(lower_limit=0.99, upper_limit=1.01)
    bulge.effective_radius = af.UniformPrior(lower_limit=1.25, upper_limit=1.35)

    phase1 = ag.PhaseImaging(
        phase_name="phase_1",
        folders=setup.folders,
        galaxies=dict(
            galaxy=ag.GalaxyModel(redshift=0.5, bulge=ag.lp.EllipticalSersic)
        ),
        search=search,
    )

    phase1.search.const_efficiency_mode = True
    phase1.search.n_live_points = 40
    phase1.search.sampling_efficiency = 0.8

    class GridPhase(af.as_grid_search(phase_class=ag.PhaseImaging, parallel=False)):
        @property
        def grid_priors(self):
            return [self.model.galaxies.light.bulge.effective_radius]

    phase2 = GridPhase(
        phase_name="phase_2",
        folders=setup.folders,
        galaxies=dict(galaxy=ag.GalaxyModel(redshift=0.5, bulge=bulge)),
        number_of_steps=2,
        search=search,
    )

    phase2.search.const_efficiency_mode = True
    phase2.search.n_live_points = 10
    phase2.search.sampling_efficiency = 0.5

    phase3 = ag.PhaseImaging(
        phase_name="phase_3",
        folders=setup.folders,
        galaxies=dict(galaxy=phase2.result.model.galaxies.light),
        search=search,
    )

    phase3.search.const_efficiency_mode = True
    phase3.search.n_live_points = 40
    phase3.search.sampling_efficiency = 0.8

    return ag.PipelineDataset(name, phase1, phase2, phase3)


if __name__ == "__main__":
    import sys

    runner.run(sys.modules[__name__])



class GalaxyData:
    def __init__(self, image, noise_map, pixel_scales):
        """ A galaxy-fit data is a collection of fit data components which are used to fit a galaxy to another galaxy. \
        This is where a component of a galaxy's light profiles (e.g. image) or mass profiles (e.g. convergence \
        , potential or deflection angles) are fitted to one another.

        This is primarily performed for automatic prior linking, as a means to efficiently link the priors of a galaxy \
        using one inferred parametrization of light or mass profiles to a new galaxy with a different parametrization \
        of light or mass profiles.

        This omits a number of the fit data components typically used when fitting an image (e.g. the observed image, PSF, \
        exposure time map), but still has a number of the other components (e.g. an effective noise_map, grid_stacks).

        Parameters
        ----------
        image : aa.Array
            An image of the quantity of the galaxy that is being fitted (e.g. its image, convergence, etc.).
        noise_map : aa.Scaled
            The noise_map used for computing the log likelihood of each fit. This can be chosen arbritarily.
        """
        self.image = image
        self.noise_map = noise_map
        self.pixel_scales = pixel_scales



from test_autogalaxy.integration.tests.imaging.galaxy_x1 import (
    galaxy_x2__sersics__separate,
)
from test_autogalaxy.integration.tests.imaging.runner import run_a_mock


class TestCase:
    def _test__galaxy_x2__sersics__separate(self):
        run_a_mock(galaxy_x2__sersics__separate)



import copy

import numpy as np
from autoarray.structures import grids
from autoarray.dataset import imaging
from autogalaxy import exc
from autogalaxy.plane import plane as pl


class MaskedImaging(imaging.MaskedImaging):
    def __init__(
        self,
        imaging,
        mask,
        grid_class=grids.Grid,
        grid_inversion_class=grids.Grid,
        fractional_accuracy=0.9999,
        sub_steps=None,
        psf_shape_2d=None,
        inversion_pixel_limit=None,
        renormalize_psf=True,
    ):
        """
        The lens dataset is the collection of data (image, noise map, PSF), a mask, grid, convolver \
        and other utilities that are used for modeling and fitting an image of a strong lens.

        Whilst the image, noise map, etc. are loaded in 2D, the lens dataset creates reduced 1D arrays of each \
        for lens calculations.

        Parameters
        ----------
        imaging: im.Imaging
            The imaging data all in 2D (the image, noise map, PSF, etc.)
        mask: msk.Mask
            The 2D mask that is applied to the image.
        sub_size : int
            The size of the sub-grid used for each lens SubGrid. E.g. a value of 2 grid each image-pixel on a 2x2 \
            sub-grid.
        psf_shape_2d : (int, int)
            The shape of the PSF used for convolving model image generated using analytic light profiles. A smaller \
            shape will trim the PSF relative to the input image PSF, giving a faster analysis run-time.
        positions : [[]]
            Lists of image-pixel coordinates (arc-seconds) that mappers close to one another in the source-plane(s), \
            used to speed up the non-linear sampling.
        pixel_scales_interp : float
            If *True*, expensive to compute mass profile deflection angles will be computed on a sparse grid and \
            interpolated to the grid, sub and blurring grids.
        inversion_pixel_limit : int or None
            The maximum number of pixels that can be used by an inversion, with the limit placed primarily to speed \
            up run.
        """

        super(MaskedImaging, self).__init__(
            imaging=imaging,
            mask=mask,
            grid_class=grid_class,
            grid_inversion_class=grid_inversion_class,
            fractional_accuracy=fractional_accuracy,
            sub_steps=sub_steps,
            psf_shape_2d=psf_shape_2d,
            inversion_pixel_limit=inversion_pixel_limit,
            renormalize_psf=renormalize_psf,
        )

    def check_inversion_pixels_are_below_limit_via_plane(self, plane):

        if self.inversion_pixel_limit is not None:
            if plane.has_pixelization:
                if plane.pixelization.pixels > self.inversion_pixel_limit:
                    raise exc.PixelizationException


class SimulatorImaging(imaging.SimulatorImaging):
    def __init__(
        self,
        psf,
        exposure_time_map,
        background_sky_map=None,
        renormalize_psf=True,
        add_noise=True,
        noise_if_add_noise_false=0.1,
        noise_seed=-1,
    ):
        """A class representing a Imaging observation, using the shape of the image, the pixel scale,
        psf, exposure time, etc.

        Parameters
        ----------
        shape_2d : (int, int)
            The shape of the observation. Note that we do not simulator a full Imaging frame (e.g. 2000 x 2000 pixels for \
            Hubble imaging), but instead just a cut-out around the strong lens.
        pixel_scales : float
            The size of each pixel in arc seconds.
        psf : PSF
            An arrays describing the PSF kernel of the image.
        exposure_time_map : float
            The exposure time of an observation using this data.
        background_sky_map : float
            The level of the background sky of an observationg using this data.
        """

        super(SimulatorImaging, self).__init__(
            psf=psf,
            exposure_time_map=exposure_time_map,
            background_sky_map=background_sky_map,
            renormalize_psf=renormalize_psf,
            add_noise=add_noise,
            noise_if_add_noise_false=noise_if_add_noise_false,
            noise_seed=noise_seed,
        )

    def from_plane_and_grid(self, plane, grid, name=None):
        """
        Create a realistic simulated image by applying effects to a plain simulated image.

        Parameters
        ----------
        name
        image : ndarray
            The image before simulating (e.g. the lens and source galaxies before optics blurring and Imaging read-out).
        pixel_scales: float
            The scale of each pixel in arc seconds
        exposure_time_map : ndarray
            An arrays representing the effective exposure time of each pixel.
        psf: PSF
            An arrays describing the PSF the simulated image is blurred with.
        background_sky_map : ndarray
            The value of background sky in every image pixel (electrons per second).
        add_noise: Bool
            If True poisson noise_maps is simulated and added to the image, based on the total counts in each image
            pixel
        noise_seed: int
            A seed for random noise_maps generation
        """

        image = plane.padded_image_from_grid_and_psf_shape(
            grid=grid, psf_shape_2d=self.psf.shape_2d
        )

        if self.psf is not None:

            simulator = copy.deepcopy(self)
            simulator.exposure_time_map = self.exposure_time_map.padded_from_kernel_shape(
                kernel_shape_2d=self.psf.shape_2d
            )

            if self.background_sky_map is not None:

                simulator.background_sky_map = self.background_sky_map.padded_from_kernel_shape(
                    kernel_shape_2d=self.psf.shape_2d
                )

        else:

            simulator = self

        return simulator.from_image(image=image.in_1d_binned, name=name)

    def from_galaxies_and_grid(self, galaxies, grid, name=None):
        """Simulate imaging data for this data, as follows:

        1)  Setup the image-plane grid of the Imaging arrays, which defines the coordinates used for the ray-tracing.

        2) Use this grid and the lens and source galaxies to setup a plane, which generates the image of \
           the simulated imaging data.

        3) Simulate the imaging data, using a special image which ensures edge-effects don't
           degrade simulator of the telescope optics (e.g. the PSF convolution).

        4) Plot the image using Matplotlib, if the plot_imaging bool is True.

        5) Output the dataset to .fits format if a dataset_path and data_name are specified. Otherwise, return the simulated \
           imaging data instance."""

        plane = pl.Plane(
            redshift=float(np.mean([galaxy.redshift for galaxy in galaxies])),
            galaxies=galaxies,
        )

        return self.from_plane_and_grid(plane=plane, grid=grid, name=name)



import autogalaxy as ag


class TestTags:
    def test__grids__sub_size_tags(self):

        settings = ag.PhaseSettingsImaging(grid_class=ag.GridIterate, sub_size=1)
        assert settings.grid_sub_size_tag == ""
        settings = ag.PhaseSettingsImaging(grid_class=ag.Grid, sub_size=1)
        assert settings.grid_sub_size_tag == "sub_1"
        settings = ag.PhaseSettingsImaging(grid_class=ag.Grid, sub_size=2)
        assert settings.grid_sub_size_tag == "sub_2"
        settings = ag.PhaseSettingsImaging(grid_class=ag.Grid, sub_size=4)
        assert settings.grid_sub_size_tag == "sub_4"

        settings = ag.PhaseSettingsImaging(
            grid_inversion_class=ag.GridIterate, sub_size=1
        )
        assert settings.grid_inversion_sub_size_tag == ""
        settings = ag.PhaseSettingsImaging(grid_inversion_class=ag.Grid, sub_size=1)
        assert settings.grid_inversion_sub_size_tag == "sub_1"
        settings = ag.PhaseSettingsImaging(grid_inversion_class=ag.Grid, sub_size=2)
        assert settings.grid_inversion_sub_size_tag == "sub_2"
        settings = ag.PhaseSettingsImaging(grid_inversion_class=ag.Grid, sub_size=4)
        assert settings.grid_inversion_sub_size_tag == "sub_4"

    def test__grids__fractional_accuracy_tags(self):

        settings = ag.PhaseSettingsImaging(grid_class=ag.Grid, fractional_accuracy=1)
        assert settings.grid_fractional_accuracy_tag == ""
        settings = ag.PhaseSettingsImaging(
            grid_class=ag.GridIterate, fractional_accuracy=0.5
        )
        assert settings.grid_fractional_accuracy_tag == "facc_0.5"
        settings = ag.PhaseSettingsImaging(
            grid_class=ag.GridIterate, fractional_accuracy=0.71
        )
        assert settings.grid_fractional_accuracy_tag == "facc_0.71"
        settings = ag.PhaseSettingsImaging(
            grid_class=ag.GridIterate, fractional_accuracy=0.999999
        )
        assert settings.grid_fractional_accuracy_tag == "facc_0.999999"

        settings = ag.PhaseSettingsImaging(
            grid_inversion_class=ag.Grid, fractional_accuracy=1
        )
        assert settings.grid_inversion_fractional_accuracy_tag == ""
        settings = ag.PhaseSettingsImaging(
            grid_inversion_class=ag.GridIterate, fractional_accuracy=0.5
        )
        assert settings.grid_inversion_fractional_accuracy_tag == "facc_0.5"
        settings = ag.PhaseSettingsImaging(
            grid_inversion_class=ag.GridIterate, fractional_accuracy=0.71
        )
        assert settings.grid_inversion_fractional_accuracy_tag == "facc_0.71"
        settings = ag.PhaseSettingsImaging(
            grid_inversion_class=ag.GridIterate, fractional_accuracy=0.999999
        )
        assert settings.grid_inversion_fractional_accuracy_tag == "facc_0.999999"

    def test__grid__pixel_scales_interp_tag(self):

        settings = ag.PhaseSettingsImaging(grid_class=ag.Grid, pixel_scales_interp=0.1)
        assert settings.grid_pixel_scales_interp_tag == ""
        settings = ag.PhaseSettingsImaging(
            grid_class=ag.GridInterpolate, pixel_scales_interp=None
        )
        assert settings.grid_pixel_scales_interp_tag == ""
        settings = ag.PhaseSettingsImaging(
            grid_class=ag.GridInterpolate, pixel_scales_interp=0.5
        )
        assert settings.grid_pixel_scales_interp_tag == "interp_0.500"
        settings = ag.PhaseSettingsImaging(
            grid_class=ag.GridInterpolate, pixel_scales_interp=0.25
        )
        assert settings.grid_pixel_scales_interp_tag == "interp_0.250"
        settings = ag.PhaseSettingsImaging(
            grid_class=ag.GridInterpolate, pixel_scales_interp=0.234
        )
        assert settings.grid_pixel_scales_interp_tag == "interp_0.234"

        settings = ag.PhaseSettingsImaging(
            grid_inversion_class=ag.Grid, pixel_scales_interp=0.1
        )
        assert settings.grid_inversion_pixel_scales_interp_tag == ""
        settings = ag.PhaseSettingsImaging(
            grid_inversion_class=ag.GridInterpolate, pixel_scales_interp=None
        )
        assert settings.grid_inversion_pixel_scales_interp_tag == ""
        settings = ag.PhaseSettingsImaging(
            grid_inversion_class=ag.GridInterpolate, pixel_scales_interp=0.5
        )
        assert settings.grid_inversion_pixel_scales_interp_tag == "interp_0.500"
        settings = ag.PhaseSettingsImaging(
            grid_inversion_class=ag.GridInterpolate, pixel_scales_interp=0.25
        )
        assert settings.grid_inversion_pixel_scales_interp_tag == "interp_0.250"
        settings = ag.PhaseSettingsImaging(
            grid_inversion_class=ag.GridInterpolate, pixel_scales_interp=0.234
        )
        assert settings.grid_inversion_pixel_scales_interp_tag == "interp_0.234"

    def test__grid_tags(self):

        settings = ag.PhaseSettingsImaging(
            grid_class=ag.Grid,
            sub_size=1,
            grid_inversion_class=ag.GridIterate,
            fractional_accuracy=0.5,
        )
        assert settings.grid_no_inversion_tag == "__grid_sub_1"
        assert settings.grid_with_inversion_tag == "__grid_sub_1_inv_facc_0.5"

        settings = ag.PhaseSettingsImaging(
            grid_class=ag.GridInterpolate,
            grid_inversion_class=ag.GridInterpolate,
            pixel_scales_interp=0.5,
        )
        assert settings.grid_no_inversion_tag == "__grid_interp_0.500"
        assert (
            settings.grid_with_inversion_tag == "__grid_interp_0.500_inv_interp_0.500"
        )

        settings = ag.PhaseSettingsImaging(
            grid_class=ag.GridIterate,
            fractional_accuracy=0.8,
            grid_inversion_class=ag.Grid,
            sub_size=2,
        )
        assert settings.grid_no_inversion_tag == "__grid_facc_0.8"
        assert settings.grid_with_inversion_tag == "__grid_facc_0.8_inv_sub_2"

    def test__signal_to_noise_limit_tag(self):

        settings = ag.PhaseSettingsImaging(signal_to_noise_limit=None)
        assert settings.signal_to_noise_limit_tag == ""
        settings = ag.PhaseSettingsImaging(signal_to_noise_limit=1)
        assert settings.signal_to_noise_limit_tag == "__snr_1"
        settings = ag.PhaseSettingsImaging(signal_to_noise_limit=2)
        assert settings.signal_to_noise_limit_tag == "__snr_2"

    def test__bin_up_factor_tag(self):

        settings = ag.PhaseSettingsImaging(bin_up_factor=None)
        assert settings.bin_up_factor_tag == ""
        settings = ag.PhaseSettingsImaging(bin_up_factor=1)
        assert settings.bin_up_factor_tag == ""
        settings = ag.PhaseSettingsImaging(bin_up_factor=2)
        assert settings.bin_up_factor_tag == "__bin_2"

    def test__psf_shape_2d_tag(self):

        settings = ag.PhaseSettingsImaging(psf_shape_2d=None)
        assert settings.psf_shape_tag == ""
        settings = ag.PhaseSettingsImaging(psf_shape_2d=(2, 2))
        assert settings.psf_shape_tag == "__psf_2x2"
        settings = ag.PhaseSettingsImaging(psf_shape_2d=(3, 4))
        assert settings.psf_shape_tag == "__psf_3x4"

    def test__transformer_tag(self):
        settings = ag.PhaseSettingsInterferometer(transformer_class=ag.TransformerDFT)
        assert settings.transformer_tag == "__dft"
        settings = ag.PhaseSettingsInterferometer(transformer_class=ag.TransformerFFT)
        assert settings.transformer_tag == "__fft"
        settings = ag.PhaseSettingsInterferometer(transformer_class=ag.TransformerNUFFT)
        assert settings.transformer_tag == "__nufft"
        settings = ag.PhaseSettingsInterferometer(transformer_class=None)
        assert settings.transformer_tag == ""

    def test__primary_beam_shape_2d_tag(self):
        settings = ag.PhaseSettingsInterferometer(primary_beam_shape_2d=None)
        assert settings.primary_beam_shape_tag == ""
        settings = ag.PhaseSettingsInterferometer(primary_beam_shape_2d=(2, 2))
        assert settings.primary_beam_shape_tag == "__pb_2x2"
        settings = ag.PhaseSettingsInterferometer(primary_beam_shape_2d=(3, 4))
        assert settings.primary_beam_shape_tag == "__pb_3x4"

    def test__tag__mixture_of_values(self):

        settings = ag.PhaseSettingsImaging(
            grid_class=ag.Grid,
            grid_inversion_class=ag.Grid,
            sub_size=2,
            signal_to_noise_limit=2,
            bin_up_factor=None,
            psf_shape_2d=None,
        )

        assert settings.phase_no_inversion_tag == "settings__grid_sub_2__snr_2"
        assert (
            settings.phase_with_inversion_tag == "settings__grid_sub_2_inv_sub_2__snr_2"
        )

        settings = ag.PhaseSettingsImaging(
            grid_class=ag.Grid,
            grid_inversion_class=ag.GridIterate,
            sub_size=1,
            fractional_accuracy=0.1,
            signal_to_noise_limit=None,
            bin_up_factor=3,
            psf_shape_2d=(2, 2),
        )

        assert settings.phase_no_inversion_tag == "settings__grid_sub_1__bin_3__psf_2x2"
        assert (
            settings.phase_with_inversion_tag
            == "settings__grid_sub_1_inv_facc_0.1__bin_3__psf_2x2"
        )

        settings = ag.PhaseSettingsInterferometer(
            grid_class=ag.GridIterate,
            grid_inversion_class=ag.Grid,
            fractional_accuracy=0.1,
            sub_size=3,
            transformer_class=ag.TransformerDFT,
            primary_beam_shape_2d=(2, 2),
            log_likelihood_cap=200.001,
        )

        assert (
            settings.phase_no_inversion_tag
            == "settings__grid_facc_0.1__dft__pb_2x2__lh_cap_200.0"
        )
        assert (
            settings.phase_with_inversion_tag
            == "settings__grid_facc_0.1_inv_sub_3__dft__pb_2x2__lh_cap_200.0"
        )



import autofit as af
from autogalaxy.galaxy import galaxy as g


class Result(af.Result):
    def __init__(
        self, samples, previous_model, analysis, search, use_as_hyper_dataset=False
    ):
        """
        The results of a non-linear search performed by a phase.

        Parameters
        ----------
        samples : af.Samples
            A class containing the samples of the non-linear search, including methods to get the maximum log
            likelihood model, errors, etc.
        previous_model : af.ModelMapper
            The model used in this result model-fit.
        analysis : Analysis
            The Analysis class used by this model-fit to fit the model to the data.
        search : af.NonLinearSearch
            The non-linear search search used by this model fit.
        use_as_hyper_dataset : bool
            Whether this result's phase contains hyper phases, allowing it to be used a hyper dataset.
        """
        super().__init__(samples=samples, previous_model=previous_model)

        self.analysis = analysis
        self.search = search
        self.use_as_hyper_dataset = use_as_hyper_dataset

    @property
    def max_log_likelihood_plane(self):

        instance = self.analysis.associate_hyper_images(instance=self.instance)

        return self.analysis.plane_for_instance(instance=instance)

    @property
    def path_galaxy_tuples(self) -> [(str, g.Galaxy)]:
        """
        Tuples associating the names of galaxies with instances from the best fit
        """
        return self.instance.path_instance_tuples_for_class(cls=g.Galaxy)



from autoconf import conf
import autofit as af


class PipelineSetup:
    def __init__(
        self,
        folders=None,
        hyper_galaxies=False,
        hyper_image_sky=False,
        hyper_background_noise=False,
        hyper_galaxy_phase_first=False,
        hyper_galaxies_search=None,
        inversion_search=None,
        hyper_combined_search=None,
        pixelization=None,
        regularization=None,
        light_centre=None,
        align_bulge_disk_centre=False,
        align_bulge_disk_elliptical_comps=False,
        disk_as_sersic=False,
        number_of_gaussians=None,
    ):
        """The setup of a pipeline, which controls how PyAutoGalaxy template pipelines runs, for example controlling
        assumptions about the bulge-disk model or the number of Gaussians used for multi-Gaussian fitting.

        Users can write their own pipelines which do not use or require the *PipelineSetup* class.

        This class enables pipeline tagging, whereby the setup of the pipeline is used in the template pipeline
        scripts to tag the output path of the results depending on the setup parameters. This allows one to fit
        different models to a dataset in a structured path format.

        Parameters
        ----------
        hyper_galaxies : bool
            If a hyper-pipeline is being used, this determines if hyper-galaxy functionality is used to scale the
            noise-map of the dataset throughout the fitting.
        hyper_image_sky : bool
            If a hyper-pipeline is being used, this determines if hyper-galaxy functionality is used include the
            image's background sky component in the model.
        hyper_background_noise : bool
            If a hyper-pipeline is being used, this determines if hyper-galaxy functionality is used include the
            noise-map's background component in the model.
        hyper_galaxies_search : af.NonLinearSearch or None
            The non-linear search used by every hyper-galaxies phase.
        inversion_search : af.NonLinearSearch or None
            The non-linear search used by every inversion phase.
        hyper_combined_search : af.NonLinearSearch or None
            The non-linear search used by every hyper combined phase.
        pixelization : ag.pix.Pixelization
           If the pipeline uses an *Inversion* to reconstruct the galaxy's light, this determines the
           *Pixelization* used.
        regularization : ag.reg.Regularization
           If the pipeline uses an *Inversion* to reconstruct the galaxy's light, this determines the
           *Regularization* scheme used.
        light_centre : (float, float)
           If input, a fixed (y,x) centre of the galaxy is used for the light profile model which is not treated as a
            free parameter by the non-linear search.
        align_bulge_disk_centre : bool
            If a bulge + disk light model (e.g. EllipticalSersic + EllipticalExponential) is used to fit the galaxy,
            *True* will align the centre of the bulge and disk components and not fit them separately.
        align_bulge_disk_phi : bool
            If a bulge + disk light model (e.g. EllipticalSersic + EllipticalExponential) is used to fit the galaxy,
            *True* will align the rotation angles phi of the bulge and disk components and not fit them separately.
        align_bulge_disk_axis_ratio : bool
            If a bulge + disk light model (e.g. EllipticalSersic + EllipticalExponential) is used to fit the galaxy,
            *True* will align the axis-ratios of the bulge and disk components and not fit them separately.
        disk_as_sersic : bool
            If a bulge + disk light model (e.g. EllipticalSersic + EllipticalExponential) is used to fit the galaxy,
            *True* will use an EllipticalSersic for the disk instead of an EllipticalExponential.
        number_of_gaussians : int
            If a multi-Gaussian light model is used to fit the galaxy, this determines the number of Gaussians.
        """

        self.folders = folders

        self.hyper_galaxies = hyper_galaxies

        if self.hyper_galaxies and hyper_galaxies_search is None:
            self.hyper_galaxies_search = af.DynestyStatic(
                n_live_points=75, sampling_efficiency=0.5, evidence_tolerance=0.8
            )
        elif self.hyper_galaxies and hyper_galaxies_search is not None:
            self.hyper_galaxies_search = hyper_galaxies_search
        else:
            self.hyper_galaxies_search = None

        if inversion_search is None:
            self.inversion_search = af.DynestyStatic(
                n_live_points=30, sampling_efficiency=0.5, evidence_tolerance=0.8
            )
        elif inversion_search is not None:
            self.inversion_search = inversion_search

        if hyper_combined_search is None:
            self.hyper_combined_search = af.DynestyStatic(
                n_live_points=50, sampling_efficiency=0.5, evidence_tolerance=0.8
            )
        else:
            self.hyper_combined_search = hyper_combined_search

        self.hyper_galaxy_phase_first = hyper_galaxy_phase_first

        self.hyper_image_sky = hyper_image_sky
        self.hyper_background_noise = hyper_background_noise

        self.pixelization = pixelization
        self.regularization = regularization
        self.light_centre = light_centre

        self.align_bulge_disk_centre = align_bulge_disk_centre
        self.align_bulge_disk_elliptical_comps = align_bulge_disk_elliptical_comps
        self.disk_as_sersic = disk_as_sersic
        self.number_of_gaussians = number_of_gaussians

    @property
    def hyper_tag(self):
        """Tag ithe hyper pipeline features used in a hyper pipeline to customize pipeline output paths.
        """
        if not any(
            [self.hyper_galaxies, self.hyper_image_sky, self.hyper_background_noise]
        ):
            return ""

        return (
            "__"
            + conf.instance.tag.get("pipeline", "hyper", str)
            + self.hyper_galaxies_tag
            + self.hyper_image_sky_tag
            + self.hyper_background_noise_tag
        )

    @property
    def hyper_galaxies_tag(self):
        """Tag if hyper-galaxies are used in a hyper pipeline to customize pipeline output paths.

        This is used to generate an overall hyper tag in *hyper_tag*.
        """
        if not self.hyper_galaxies:
            return ""
        elif self.hyper_galaxies:
            return "_" + conf.instance.tag.get("pipeline", "hyper_galaxies", str)

    @property
    def hyper_image_sky_tag(self):
        """Tag if the sky-background is hyper as a hyper_galaxies-parameter in a hyper pipeline to customize pipeline
        output paths.

        This is used to generate an overall hyper tag in *hyper_tag*.
        """
        if not self.hyper_image_sky:
            return ""
        elif self.hyper_image_sky:
            return "_" + conf.instance.tag.get("pipeline", "hyper_image_sky", str)

    @property
    def hyper_background_noise_tag(self):
        """Tag if the background noise is hyper as a hyper_galaxies-parameter in a hyper pipeline to customize pipeline
        output paths.

        This is used to generate an overall hyper tag in *hyper_tag*.
        """
        if not self.hyper_background_noise:
            return ""
        elif self.hyper_background_noise:
            return "_" + conf.instance.tag.get(
                "pipeline", "hyper_background_noise", str
            )

    @property
    def tag(self):
        """Generate the pipeline's overall tag, which customizes the 'setup' folder the results are output to.
        """
        return (
            conf.instance.tag.get("pipeline", "pipeline", str)
            + self.hyper_tag
            + self.inversion_tag
            + self.light_centre_tag
            + self.align_bulge_disk_tag
            + self.disk_as_sersic_tag
            + self.number_of_gaussians_tag
        )

    @property
    def inversion_tag(self):
        """Generate a tag if an *Inversion* is used to  *Pixelization* used to reconstruct the galaxy's light, which 
        is the sum of the pixelization and regularization tags.
        """
        if self.pixelization is None or self.regularization is None:
            return ""

        return "__" + self.pixelization_tag + self.regularization_tag

    @property
    def inversion_tag_no_underscore(self):
        """Generate a tag if an *Inversion* is used to  *Pixelization* used to reconstruct the galaxy's light, which
        is the sum of the pixelization and regularization tags.
        """
        if self.pixelization is None or self.regularization is None:
            return ""

        return self.pixelization_tag + self.regularization_tag

    @property
    def pixelization_tag(self):
        """Tag the *Pixelization* scheme used by the pipeline, if an inversion is usse to reconstruct the galaxy's 
        light.

        The pixelization tag is loaded from the config file 'autogalaxy_workspace/config/label.ini' in the [tag]
        seciton.

        For the default regulariation tags, this changes the setup folder as follows:

        pixelization = None -> setup
        pixelization = pix.Rectangular -> setup__pix_rect
        pixelization = pix.VoronoiMagnification -> setup__pix_voro_mag
        """
        if self.pixelization is None:
            return ""
        else:
            return (
                conf.instance.tag.get("pipeline", "pixelization", str)
                + "_"
                + conf.instance.tag.get(
                    "pixelization", self.pixelization().__class__.__name__, str
                )
            )

    @property
    def regularization_tag(self):
        """Tag the *Regularization* scheme used by the pipeline, if an inversion is usse to reconstruct the galaxy's 
        light.
        
        The regularization tag is loaded from the config file 'autogalaxy_workspace/config/label.ini' in the [tag]
        seciton.

        For the default regulariation tags, this changes the setup folder as follows:

        regularization = None -> setup
        regularization = reg.Constant -> setup__reg_const
        regularization = reg.AdaptiveBrightness -> setup__reg_adapt_bright
        """
        if self.regularization is None:
            return ""
        else:
            return (
                "__"
                + conf.instance.tag.get("pipeline", "regularization", str)
                + "_"
                + conf.instance.tag.get(
                    "regularization", self.regularization().__class__.__name__, str
                )
            )

    @property
    def light_centre_tag(self):
        """Tag if the lens light of the pipeline are fixed to a previous estimate, or varied \
         during the analysis, to customize pipeline output paths.

        This changes the setup folder as follows:

        light_centre = None -> setup
        light_centre = (1.0, 1.0) -> setup___light_centre_(1.0, 1.0)
        light_centre = (3.0, -2.0) -> setup___light_centre_(3.0, -2.0)
        """
        if self.light_centre is None:
            return ""
        else:
            y = "{0:.2f}".format(self.light_centre[0])
            x = "{0:.2f}".format(self.light_centre[1])
            return (
                "__"
                + conf.instance.tag.get("pipeline", "light_centre", str)
                + "_("
                + y
                + ","
                + x
                + ")"
            )

    @property
    def align_bulge_disk_centre_tag(self):
        """Tag if the bulge and disk of a bulge-disk system are aligned or not, to customize pipeline 
        output paths based on the bulge-disk model.
        
        This is used to generate an overall align-bulge disk tag in *align_bulge_disk_tag*.
        """
        if not self.align_bulge_disk_centre:
            return ""
        elif self.align_bulge_disk_centre:
            return "_" + conf.instance.tag.get(
                "pipeline", "align_bulge_disk_centre", str
            )

    @property
    def align_bulge_disk_elliptical_comps_tag(self):
        """Tag if the ellipticity of the bulge and disk of a bulge-disk system are aligned or not, to customize pipeline
        output paths based on the bulge-disk model. 
        
        This is used to generate an overall align-bulge disk tag in *align_bulge_disk_tag*.
        """
        if not self.align_bulge_disk_elliptical_comps:
            return ""
        elif self.align_bulge_disk_elliptical_comps:
            return "_" + conf.instance.tag.get(
                "pipeline", "align_bulge_disk_elliptical_comps", str
            )

    @property
    def align_bulge_disk_tag(self):
        """Tag the alignment of the geometry of the bulge and disk of a bulge-disk system, to customize  pipeline output 
        paths based on the bulge-disk model. This adds together the bulge_disk tags generated in the  3 functions 
        *align_bulge_disk_centre_tag*, *align_bulge_disk_axis_ratio_tag* and *align_bulge_disk_phi_tag*.
        """

        if not any(
            [self.align_bulge_disk_centre, self.align_bulge_disk_elliptical_comps]
        ):
            return ""

        return (
            "__"
            + conf.instance.tag.get("pipeline", "align_bulge_disk", str)
            + self.align_bulge_disk_centre_tag
            + self.align_bulge_disk_elliptical_comps_tag
        )

    @property
    def disk_as_sersic_tag(self):
        """Tag if the disk component of a bulge-disk light profile fit of the pipeline is modeled as a EllipticalSersic
        or an EllipticalExponential.

        This changes the setup folder as follows:

        disk_as_sersic = False -> setup
        disk_as_sersic = True -> setup___disk_as_sersic
        """
        if not self.disk_as_sersic:
            return ""
        elif self.disk_as_sersic:
            return "__" + conf.instance.tag.get("pipeline", "disk_as_sersic", str)

    @property
    def number_of_gaussians_tag(self):
        """Tag the number of Gaussians if modeling the galaxy using multiple Gaussians light profiles.

        This changes the setup folder as follows:

        number_of_gaussians = None -> setup
        number_of_gaussians = 2 = True -> setup__gaussians_x2
        number_of_gaussians = 4 = True -> setup__gaussians_x4
        """
        if self.number_of_gaussians is None:
            return ""
        else:
            return (
                "__"
                + conf.instance.tag.get("pipeline", "number_of_gaussians", str)
                + "_x"
                + str(self.number_of_gaussians)
            )



import os
from test import integration_util

import autofit as af
import autogalaxy as ag
import numpy as np

test_type = "galaxy_fit"
test_name = "image"

test_path = "{}/../../".format(os.path.dirname(os.path.realpath(__file__)))
output_path = test_path + "output/"


def galaxy_fit_phase():

    pixel_scales = 0.1
    image_shape = (150, 150)

    integration_util.reset_paths(test_name=test_name, output_path=output_path)

    grid = ag.Grid.uniform(shape_2d=image_shape, pixel_scales=pixel_scales, sub_size=4)

    galaxy = ag.Galaxy(
        redshift=0.5,
        light=ag.lp.SphericalExponential(
            centre=(0.0, 0.0), intensity=1.0, effective_radius=0.5
        ),
    )

    image = galaxy.image_from_grid(galaxies=[galaxy], grid=grid)

    noise_map = ag.Array.manual_2d(
        array=np.ones(image.shape), pixel_scales=pixel_scales
    )

    data = ag.GalaxyData(image=image, noise_map=noise_map, pixel_scales=pixel_scales)

    phase1 = ag.PhaseGalaxy(
        phase_name=test_name + "/",
        galaxies=dict(
            gal=ag.GalaxyModel(redshift=0.5, light=ag.lp.SphericalExponential)
        ),
        use_image=True,
        sub_size=4,
        search=af.DynestyStatic(),
    )

    phase1.run(galaxy_data=[data])


if __name__ == "__main__":
    galaxy_fit_phase()



import autofit as af
import autogalaxy as ag
from test_autogalaxy.integration.tests.imaging import runner

test_type = "galaxy_x1"
test_name = "galaxy_x2__sersics__separate"
data_label = "galaxy_x2__sersics"
instrument = "vro"


def make_pipeline(name, folders, search=af.DynestyStatic()):

    bulge_0 = af.PriorModel(ag.lp.EllipticalSersic)

    bulge_0.centre_0 = -1.0
    bulge_0.centre_1 = -1.0

    phase1 = ag.PhaseImaging(
        phase_name="phase_1",
        folders=setup.folders,
        galaxies=dict(galaxy_0=ag.GalaxyModel(redshift=0.5, bulge=bulge_0)),
        search=search,
    )

    phase1.search.const_efficiency_mode = True
    phase1.search.n_live_points = 40
    phase1.search.sampling_efficiency = 0.8

    bulge_1 = af.PriorModel(ag.lp.EllipticalSersic)

    bulge_1.centre_0 = 1.0
    bulge_1.centre_1 = 1.0

    phase2 = ag.PhaseImaging(
        phase_name="phase_2",
        folders=setup.folders,
        galaxies=dict(
            galaxy_0=phase1.result.instance.galaxies.galaxy_0,
            galaxy_1=ag.GalaxyModel(redshift=0.5, bulge=bulge_1),
        ),
        search=search,
    )

    phase2.search.const_efficiency_mode = True
    phase2.search.n_live_points = 40
    phase2.search.sampling_efficiency = 0.8

    phase3 = ag.PhaseImaging(
        phase_name="phase_3",
        folders=setup.folders,
        galaxies=dict(
            galaxy_0=phase1.result.model.galaxies.galaxy_0,
            galaxy_1=phase2.result.model.galaxies.galaxy_1,
        ),
        search=search,
    )

    phase3.search.const_efficiency_mode = True
    phase3.search.n_live_points = 60
    phase3.search.sampling_efficiency = 0.8

    return ag.PipelineDataset(name, phase1, phase2, phase3)


if __name__ == "__main__":
    import sys

    runner.run(sys.modules[__name__])



import copy

import autofit as af
from autofit.tools.phase import Dataset
from autogalaxy.pipeline.phase import abstract


class HyperPhase:
    def __init__(self, phase: abstract.AbstractPhase, search, hyper_name: str):
        """
        Abstract HyperPhase. Wraps a phase, performing that phase before performing the action
        specified by the run_hyper.

        Parameters
        ----------
        phase
            A phase
        """
        self.phase = phase
        self.hyper_name = hyper_name
        self.search = search

    def run_hyper(self, *args, **kwargs) -> af.Result:
        """
        Run the hyper_galaxies phase.

        Parameters
        ----------
        args
        kwargs

        Returns
        -------
        result
            The result of the hyper_galaxies phase.
        """
        raise NotImplementedError()

    def make_hyper_phase(self) -> abstract.AbstractPhase:
        """
        Returns
        -------
        hyper_phase
            A copy of the original phase with a modified name and path
        """

        phase = copy.deepcopy(self.phase)
        phase.paths.zip()
        new_output_path = f"{self.phase.paths.path_prefix}/{self.phase.phase_name}/{self.hyper_name}__{phase.paths.tag}"

        phase.search = self.search.copy_with_name_extension(
            extension=new_output_path, remove_phase_tag=True
        )

        phase.is_hyper_phase = True

        return phase

    def run(
        self, dataset: Dataset, results: af.ResultsCollection = None, **kwargs
    ) -> af.Result:
        """
        Run the hyper phase and then the hyper_galaxies phase.

        Parameters
        ----------
        dataset
            Data
        results
            Results from previous phases.
        kwargs

        Returns
        -------
        result
            The result of the phase, with a hyper_galaxies result attached as an attribute with the hyper_name of this
            phase.
        """
        self.save_dataset(dataset=dataset)

        results = (
            copy.deepcopy(results) if results is not None else af.ResultsCollection()
        )

        result = self.phase.run(dataset, results=results, **kwargs)
        results.add(self.phase.paths.name, result)
        hyper_result = self.run_hyper(dataset=dataset, results=results, **kwargs)
        setattr(result, self.hyper_name, hyper_result)
        return result

    def __getattr__(self, item):
        return getattr(self.phase, item)



from autoarray.plot import plotters
from autoarray.structures import arrays
from autogalaxy.plot import lensing_plotters


@lensing_plotters.set_include_and_plotter
@plotters.set_labels
def image(plane, grid, positions=None, include=None, plotter=None):

    plotter.plot_array(
        array=plane.image_from_grid(grid=grid),
        mask=include.mask_from_grid(grid=grid),
        positions=positions,
        critical_curves=include.critical_curves_from_obj(obj=plane),
        light_profile_centres=include.light_profile_centres_from_obj(obj=plane),
        mass_profile_centres=include.mass_profile_centres_from_obj(obj=plane),
        include_origin=include.origin,
    )


@lensing_plotters.set_include_and_plotter
@plotters.set_labels
def plane_image(plane, grid, positions=None, caustics=None, include=None, plotter=None):

    plotter.plot_array(
        array=plane.plane_image_from_grid(grid=grid).array,
        positions=positions,
        caustics=caustics,
        grid=include.grid_from_grid(grid=grid),
        light_profile_centres=include.light_profile_centres_from_obj(obj=plane),
        mass_profile_centres=include.mass_profile_centres_from_obj(obj=plane),
        include_origin=include.origin,
    )


@lensing_plotters.set_include_and_plotter
@plotters.set_labels
def convergence(plane, grid, include=None, plotter=None):

    plotter.plot_array(
        array=plane.convergence_from_grid(grid=grid),
        mask=include.mask_from_grid(grid=grid),
        critical_curves=include.critical_curves_from_obj(obj=plane),
        light_profile_centres=include.light_profile_centres_from_obj(obj=plane),
        mass_profile_centres=include.mass_profile_centres_from_obj(obj=plane),
        include_origin=include.origin,
    )


@lensing_plotters.set_include_and_plotter
@plotters.set_labels
def potential(plane, grid, include=None, plotter=None):

    plotter.plot_array(
        array=plane.potential_from_grid(grid=grid),
        mask=include.mask_from_grid(grid=grid),
        critical_curves=include.critical_curves_from_obj(obj=plane),
        light_profile_centres=include.light_profile_centres_from_obj(obj=plane),
        mass_profile_centres=include.mass_profile_centres_from_obj(obj=plane),
        include_origin=include.origin,
    )


@lensing_plotters.set_include_and_plotter
@plotters.set_labels
def deflections_y(plane, grid, include=None, plotter=None):

    deflections = plane.deflections_from_grid(grid=grid)
    deflections_y = arrays.MaskedArray.manual_1d(
        array=deflections.in_1d[:, 0], mask=grid.mask
    )

    plotter.plot_array(
        array=deflections_y,
        mask=include.mask_from_grid(grid=grid),
        critical_curves=include.critical_curves_from_obj(obj=plane),
        light_profile_centres=include.light_profile_centres_from_obj(obj=plane),
        mass_profile_centres=include.mass_profile_centres_from_obj(obj=plane),
        include_origin=include.origin,
    )


@lensing_plotters.set_include_and_plotter
@plotters.set_labels
def deflections_x(plane, grid, include=None, plotter=None):

    deflections = plane.deflections_from_grid(grid=grid)
    deflections_x = arrays.MaskedArray.manual_1d(
        array=deflections.in_1d[:, 1], mask=grid.mask
    )

    plotter.plot_array(
        array=deflections_x,
        mask=include.mask_from_grid(grid=grid),
        critical_curves=include.critical_curves_from_obj(obj=plane),
        light_profile_centres=include.light_profile_centres_from_obj(obj=plane),
        mass_profile_centres=include.mass_profile_centres_from_obj(obj=plane),
        include_origin=include.origin,
    )


@lensing_plotters.set_include_and_plotter
@plotters.set_labels
def magnification(plane, grid, include=None, plotter=None):

    plotter.plot_array(
        array=plane.magnification_from_grid(grid=grid),
        mask=include.mask_from_grid(grid=grid),
        critical_curves=include.critical_curves_from_obj(obj=plane),
        light_profile_centres=include.light_profile_centres_from_obj(obj=plane),
        mass_profile_centres=include.mass_profile_centres_from_obj(obj=plane),
        include_origin=include.origin,
    )


@lensing_plotters.set_include_and_sub_plotter
@plotters.set_labels
def image_and_source_plane_subplot(
    image_plane,
    source_plane,
    grid,
    indexes=None,
    positions=None,
    axis_limits=None,
    include=None,
    sub_plotter=None,
):

    number_subplots = 2

    sub_plotter.open_subplot_figure(number_subplots=number_subplots)

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=1)

    plane_grid(
        plane=image_plane,
        grid=grid,
        indexes=indexes,
        axis_limits=axis_limits,
        positions=positions,
        critical_curves=include.critical_curves_from_obj(obj=image_plane),
        include=include,
        plotter=sub_plotter,
    )

    source_plane_grid = image_plane.traced_grid_from_grid(grid=grid)

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=2)

    plane_grid(
        plane=source_plane,
        grid=source_plane_grid,
        indexes=indexes,
        axis_limits=axis_limits,
        positions=positions,
        caustics=include.caustics_from_obj(obj=image_plane),
        include=include,
        plotter=sub_plotter,
    )

    sub_plotter.output.subplot_to_figure()
    sub_plotter.figure.close()


@lensing_plotters.set_include_and_plotter
@plotters.set_labels
def plane_grid(
    plane,
    grid,
    indexes=None,
    axis_limits=None,
    positions=None,
    critical_curves=None,
    caustics=None,
    include=None,
    plotter=None,
):

    plotter.plot_grid(
        grid=grid,
        positions=positions,
        axis_limits=axis_limits,
        indexes=indexes,
        critical_curves=critical_curves,
        caustics=caustics,
        light_profile_centres=include.light_profile_centres_from_obj(obj=plane),
        mass_profile_centres=include.mass_profile_centres_from_obj(obj=plane),
        include_origin=include.origin,
        include_border=include.border,
    )


@lensing_plotters.set_include_and_plotter
@plotters.set_labels
def contribution_map(plane, mask=None, positions=None, include=None, plotter=None):

    plotter.plot_array(
        array=plane.contribution_map,
        mask=mask,
        positions=positions,
        light_profile_centres=include.light_profile_centres_from_obj(obj=plane),
        mass_profile_centres=include.mass_profile_centres_from_obj(obj=plane),
        critical_curves=include.critical_curves_from_obj(obj=plane),
        include_origin=include.origin,
        include_border=include.border,
    )



import math

from autoconf import conf
import autogalaxy as ag
import numpy as np
import pytest
from test_autogalaxy import mock


def mass_within_radius_of_profile_from_grid_calculation(radius, profile):

    mass_total = 0.0

    xs = np.linspace(-radius * 1.5, radius * 1.5, 40)
    ys = np.linspace(-radius * 1.5, radius * 1.5, 40)

    edge = xs[1] - xs[0]
    area = edge ** 2

    for x in xs:
        for y in ys:

            eta = profile.grid_to_elliptical_radii(grid=np.array([[x, y]]))

            if eta < radius:
                mass_total += profile.convergence_func(eta) * area

    return mass_total


class TestMassWithinCircle:
    def test__mass_in_angular_units__singular_isothermal_sphere__compare_to_analytic(
        self
    ):

        sis = ag.mp.SphericalIsothermal(einstein_radius=2.0)

        radius = ag.dim.Length(2.0, "arcsec")

        mass = sis.mass_within_circle_in_units(
            radius=radius, redshift_object=0.5, redshift_source=1.0, unit_mass="angular"
        )
        assert math.pi * sis.einstein_radius * radius == pytest.approx(mass, 1e-3)

        sis = ag.mp.SphericalIsothermal(einstein_radius=4.0)

        radius = ag.dim.Length(4.0, "arcsec")

        mass = sis.mass_within_circle_in_units(
            radius=radius, redshift_object=0.5, redshift_source=1.0, unit_mass="angular"
        )
        assert math.pi * sis.einstein_radius * radius == pytest.approx(mass, 1e-3)

    def test__mass_in_angular_units__singular_isothermal__compare_to_grid(self):

        sis = ag.mp.SphericalIsothermal(einstein_radius=2.0)

        radius = ag.dim.Length(1.0, "arcsec")

        mass_grid = mass_within_radius_of_profile_from_grid_calculation(
            radius=radius, profile=sis
        )

        mass = sis.mass_within_circle_in_units(
            radius=radius, redshift_object=0.5, redshift_source=1.0, unit_mass="angular"
        )

        assert mass_grid == pytest.approx(mass, 0.02)

    def test__radius_units_conversions__mass_profile_updates_units_and_computes_correct_mass(
        self
    ):

        cosmology = mock.MockCosmology(kpc_per_arcsec=2.0)

        # arcsec -> arcsec

        sis_arcsec = ag.mp.SphericalIsothermal(
            centre=(ag.dim.Length(0.0, "arcsec"), ag.dim.Length(0.0, "arcsec")),
            einstein_radius=ag.dim.Length(2.0, "arcsec"),
        )

        radius = ag.dim.Length(2.0, "arcsec")
        mass = sis_arcsec.mass_within_circle_in_units(
            radius=radius, redshift_object=0.5, redshift_source=1.0, unit_mass="angular"
        )
        assert math.pi * sis_arcsec.einstein_radius * radius == pytest.approx(
            mass, 1e-3
        )

        # arcsec -> kpc

        radius = ag.dim.Length(2.0, "kpc")
        mass = sis_arcsec.mass_within_circle_in_units(
            radius=radius,
            redshift_object=0.5,
            redshift_source=1.0,
            unit_mass="angular",
            cosmology=cosmology,
        )
        assert math.pi * sis_arcsec.einstein_radius * 1.0 == pytest.approx(mass, 1e-3)

        # 2.0 arcsec = 4.0 kpc, same masses.

        radius = ag.dim.Length(2.0, "arcsec")
        mass_arcsec = sis_arcsec.mass_within_circle_in_units(
            radius=radius,
            redshift_object=0.5,
            redshift_source=1.0,
            unit_mass="angular",
            cosmology=cosmology,
        )
        radius = ag.dim.Length(4.0, "kpc")
        mass_kpc = sis_arcsec.mass_within_circle_in_units(
            radius=radius,
            redshift_object=0.5,
            redshift_source=1.0,
            unit_mass="angular",
            cosmology=cosmology,
        )
        assert mass_arcsec == mass_kpc

        # kpc -> kpc

        sis_kpc = ag.mp.SphericalIsothermal(
            centre=(ag.dim.Length(0.0, "kpc"), ag.dim.Length(0.0, "kpc")),
            einstein_radius=ag.dim.Length(2.0, "kpc"),
        )

        sis_kpc.core_radius = ag.dim.Length(value=0.0, unit_length="kpc")

        radius = ag.dim.Length(2.0, "kpc")
        mass = sis_kpc.mass_within_circle_in_units(
            radius=radius,
            redshift_object=0.5,
            redshift_source=1.0,
            unit_mass="angular",
            cosmology=cosmology,
        )
        assert math.pi * sis_kpc.einstein_radius * radius == pytest.approx(mass, 1e-3)

        # kpc -> arcsec

        radius = ag.dim.Length(2.0, "arcsec")
        mass = sis_kpc.mass_within_circle_in_units(
            radius=radius,
            redshift_object=0.5,
            redshift_source=1.0,
            unit_mass="angular",
            cosmology=cosmology,
        )
        assert 2.0 * math.pi * sis_kpc.einstein_radius * radius == pytest.approx(
            mass, 1e-3
        )

        # 2.0 arcsec = 4.0 kpc, same masses.

        radius = ag.dim.Length(2.0, "arcsec")
        mass_arcsec = sis_kpc.mass_within_circle_in_units(
            radius=radius,
            redshift_object=0.5,
            redshift_source=1.0,
            unit_mass="angular",
            cosmology=cosmology,
        )
        radius = ag.dim.Length(4.0, "kpc")
        mass_kpc = sis_kpc.mass_within_circle_in_units(
            radius=radius, redshift_object=0.5, redshift_source=1.0, unit_mass="angular"
        )
        assert mass_arcsec == mass_kpc

    def test__mass_units_conversions__multiplies_by_critical_surface_density_factor(
        self
    ):

        cosmology = mock.MockCosmology(critical_surface_density=2.0)

        sis = ag.mp.SphericalIsothermal(einstein_radius=2.0)
        radius = ag.dim.Length(2.0, "arcsec")

        mass = sis.mass_within_circle_in_units(
            radius=radius,
            redshift_object=0.5,
            redshift_source=1.0,
            unit_mass="angular",
            cosmology=cosmology,
        )
        assert math.pi * sis.einstein_radius * radius == pytest.approx(mass, 1e-3)

        mass = sis.mass_within_circle_in_units(
            radius=radius,
            redshift_object=0.5,
            redshift_source=1.0,
            unit_mass="solMass",
            cosmology=cosmology,
        )
        assert 2.0 * math.pi * sis.einstein_radius * radius == pytest.approx(mass, 1e-3)

        mass = sis.mass_within_circle_in_units(
            radius=radius,
            redshift_object=0.5,
            redshift_source=1.0,
            unit_mass="solMass",
            cosmology=cosmology,
        )
        assert 2.0 * math.pi * sis.einstein_radius * radius == pytest.approx(mass, 1e-3)


class TestDensityBetweenAnnuli:
    def test__circular_annuli__sis__analyic_density_agrees(self):

        cosmology = mock.MockCosmology(kpc_per_arcsec=2.0, critical_surface_density=2.0)

        einstein_radius = 1.0
        sis_arcsec = ag.mp.SphericalIsothermal(
            centre=(0.0, 0.0), einstein_radius=einstein_radius
        )

        inner_annuli_radius = ag.dim.Length(2.0, "arcsec")
        outer_annuli_radius = ag.dim.Length(3.0, "arcsec")

        inner_mass = math.pi * einstein_radius * inner_annuli_radius
        outer_mass = math.pi * einstein_radius * outer_annuli_radius

        density_between_annuli = sis_arcsec.density_between_circular_annuli_in_angular_units(
            inner_annuli_radius=inner_annuli_radius,
            outer_annuli_radius=outer_annuli_radius,
            unit_length="arcsec",
            unit_mass="angular",
            redshift_profile=0.5,
            redshift_source=1.0,
            cosmology=cosmology,
        )

        annuli_area = (np.pi * outer_annuli_radius ** 2.0) - (
            np.pi * inner_annuli_radius ** 2.0
        )

        assert (outer_mass - inner_mass) / annuli_area == pytest.approx(
            density_between_annuli, 1e-4
        )

        density_between_annuli = sis_arcsec.density_between_circular_annuli_in_angular_units(
            inner_annuli_radius=inner_annuli_radius,
            outer_annuli_radius=outer_annuli_radius,
            unit_length="arcsec",
            unit_mass="solMass",
            redshift_profile=0.5,
            redshift_source=1.0,
            cosmology=cosmology,
        )

        annuli_area = (np.pi * outer_annuli_radius ** 2.0) - (
            np.pi * inner_annuli_radius ** 2.0
        )

        assert (2.0 * outer_mass - 2.0 * inner_mass) / annuli_area == pytest.approx(
            density_between_annuli, 1e-4
        )

        density_between_annuli = sis_arcsec.density_between_circular_annuli_in_angular_units(
            inner_annuli_radius=inner_annuli_radius,
            outer_annuli_radius=outer_annuli_radius,
            unit_length="kpc",
            unit_mass="angular",
            redshift_profile=0.5,
            redshift_source=1.0,
            cosmology=cosmology,
        )

        inner_mass = math.pi * 2.0 * einstein_radius * inner_annuli_radius
        outer_mass = math.pi * 2.0 * einstein_radius * outer_annuli_radius

        annuli_area = (np.pi * 2.0 * outer_annuli_radius ** 2.0) - (
            np.pi * 2.0 * inner_annuli_radius ** 2.0
        )

        assert (outer_mass - inner_mass) / annuli_area == pytest.approx(
            density_between_annuli, 1e-4
        )

    def test__circular_annuli__nfw_profile__compare_to_manual_mass(self):

        cosmology = mock.MockCosmology(kpc_per_arcsec=2.0, critical_surface_density=2.0)

        nfw = ag.mp.EllipticalNFW(
            centre=(0.0, 0.0), elliptical_comps=(0.111111, 0.0), kappa_s=1.0
        )

        inner_mass = nfw.mass_within_circle_in_units(
            radius=ag.dim.Length(1.0),
            redshift_object=0.5,
            redshift_source=1.0,
            unit_mass="angular",
        )

        outer_mass = nfw.mass_within_circle_in_units(
            radius=ag.dim.Length(2.0),
            redshift_object=0.5,
            redshift_source=1.0,
            unit_mass="angular",
        )

        density_between_annuli = nfw.density_between_circular_annuli_in_angular_units(
            inner_annuli_radius=ag.dim.Length(1.0),
            outer_annuli_radius=ag.dim.Length(2.0),
            unit_length="arcsec",
            unit_mass="angular",
            redshift_profile=0.5,
            redshift_source=1.0,
            cosmology=cosmology,
        )

        annuli_area = (np.pi * 2.0 ** 2.0) - (np.pi * 1.0 ** 2.0)

        assert (outer_mass - inner_mass) / annuli_area == pytest.approx(
            density_between_annuli, 1e-4
        )


class TestLensingObject:
    def test__mass_profiles__list__is_list_of_self(self):

        sis = ag.mp.SphericalIsothermal(centre=(0.0, 0.0), einstein_radius=2.0)

        assert sis.mass_profiles == [sis]

    def test__correct_einstein_mass_caclulated__means_all_innherited_methods_work(
        self,
    ):

        sis = ag.mp.SphericalIsothermal(centre=(0.0, 0.0), einstein_radius=2.0)

        assert sis.einstein_mass_in_units(unit_mass="angular") == pytest.approx(
            np.pi * 2.0 ** 2.0, 1.0e-1
        )


class TestDecorators:
    def test__grid_iterate_in__iterates_grid_result_correctly(self, gal_x1_mp):

        mask = ag.Mask.manual(
            mask=[
                [True, True, True, True, True],
                [True, False, False, False, True],
                [True, False, False, False, True],
                [True, False, False, False, True],
                [True, True, True, True, True],
            ],
            pixel_scales=(1.0, 1.0),
        )

        grid = ag.GridIterate.from_mask(
            mask=mask, fractional_accuracy=1.0, sub_steps=[2]
        )

        mass_profile = ag.mp.EllipticalIsothermal(
            centre=(0.08, 0.08), einstein_radius=1.0
        )

        deflections = mass_profile.deflections_from_grid(grid=grid)

        mask_sub_2 = mask.mask_new_sub_size_from_mask(mask=mask, sub_size=2)
        grid_sub_2 = ag.Grid.from_mask(mask=mask_sub_2)
        deflections_sub_2 = mass_profile.deflections_from_grid(
            grid=grid_sub_2
        ).in_1d_binned

        assert deflections == pytest.approx(deflections_sub_2, 1.0e-6)

        grid = ag.GridIterate.from_mask(
            mask=mask, fractional_accuracy=0.99, sub_steps=[2, 4, 8]
        )

        mass_profile = ag.mp.EllipticalIsothermal(
            centre=(0.08, 0.08), einstein_radius=1.0
        )

        deflections = mass_profile.deflections_from_grid(grid=grid)

        mask_sub_4 = mask.mask_new_sub_size_from_mask(mask=mask, sub_size=4)
        grid_sub_4 = ag.Grid.from_mask(mask=mask_sub_4)
        deflections_sub_4 = mass_profile.deflections_from_grid(
            grid=grid_sub_4
        ).in_1d_binned

        assert deflections[0, 0] == deflections_sub_4[0, 0]

        mask_sub_8 = mask.mask_new_sub_size_from_mask(mask=mask, sub_size=8)
        grid_sub_8 = ag.Grid.from_mask(mask=mask_sub_8)
        deflections_sub_8 = mass_profile.deflections_from_grid(
            grid=grid_sub_8
        ).in_1d_binned

        assert deflections[4, 0] == deflections_sub_8[4, 0]

    def test__grid_interpolate_in__convergence__interpolates_based_on_intepolate_config(
        self
    ):

        # False in interpolate.ini

        mask = ag.Mask.manual(
            mask=[
                [True, True, True, True, True],
                [True, False, False, False, True],
                [True, False, False, False, True],
                [True, False, False, False, True],
                [True, True, True, True, True],
            ],
            pixel_scales=(1.0, 1.0),
        )

        grid = ag.Grid.from_mask(mask=mask)

        grid_interpolate = ag.GridInterpolate.from_mask(
            mask=mask, pixel_scales_interp=0.1
        )

        mass_profile = ag.mp.EllipticalIsothermal(einstein_radius=1.0)

        convergence = mass_profile.convergence_from_grid(grid=grid)
        convergence_no_interpolate = mass_profile.convergence_from_grid(
            grid=grid_interpolate
        )

        assert (convergence == convergence_no_interpolate).all()

        # False in interpolate.ini

        mass_profile = ag.mp.SphericalIsothermal(einstein_radius=1.0)

        convergence = mass_profile.convergence_from_grid(grid=grid)
        convergence_interpolate = mass_profile.convergence_from_grid(
            grid=grid_interpolate
        )
        assert (convergence != convergence_interpolate).all()

        array_interp = mass_profile.convergence_from_grid(
            grid=grid_interpolate.grid_interp
        )
        interpolated_array = grid_interpolate.interpolated_array_from_array_interp(
            array_interp=array_interp
        )
        assert (convergence_interpolate == interpolated_array).all()

    def test__grid_interpolate_in__potential__interpolates_based_on_intepolate_config(
        self
    ):

        # False in interpolate.ini

        mask = ag.Mask.manual(
            mask=[
                [True, True, True, True, True],
                [True, False, False, False, True],
                [True, False, False, False, True],
                [True, False, False, False, True],
                [True, True, True, True, True],
            ],
            pixel_scales=(1.0, 1.0),
        )

        grid = ag.Grid.from_mask(mask=mask)

        grid_interpolate = ag.GridInterpolate.from_mask(
            mask=mask, pixel_scales_interp=0.1
        )

        mass_profile = ag.mp.EllipticalIsothermal(einstein_radius=1.0)

        potential = mass_profile.potential_from_grid(grid=grid)
        potential_no_interpolate = mass_profile.potential_from_grid(
            grid=grid_interpolate
        )

        assert (potential == potential_no_interpolate).all()

        # False in interpolate.ini

        mass_profile = ag.mp.SphericalIsothermal(einstein_radius=1.0)

        potential = mass_profile.potential_from_grid(grid=grid)
        potential_interpolate = mass_profile.potential_from_grid(grid=grid_interpolate)
        assert (potential != potential_interpolate).all()

        array_interp = mass_profile.potential_from_grid(
            grid=grid_interpolate.grid_interp
        )
        interpolated_array = grid_interpolate.interpolated_array_from_array_interp(
            array_interp=array_interp
        )
        assert (potential_interpolate == interpolated_array).all()

    def test__grid_interpolate_in__deflections__interpolates_based_on_intepolate_config(
        self
    ):

        # False in interpolate.ini

        mask = ag.Mask.manual(
            mask=[
                [True, True, True, True, True],
                [True, False, False, False, True],
                [True, False, False, False, True],
                [True, False, False, False, True],
                [True, True, True, True, True],
            ],
            pixel_scales=(1.0, 1.0),
        )

        grid = ag.Grid.from_mask(mask=mask)

        grid_interpolate = ag.GridInterpolate.from_mask(
            mask=mask, pixel_scales_interp=0.1
        )

        mass_profile = ag.mp.EllipticalIsothermal(einstein_radius=1.0)

        deflections = mass_profile.deflections_from_grid(grid=grid)
        deflections_no_interpolate = mass_profile.deflections_from_grid(
            grid=grid_interpolate
        )

        assert (deflections == deflections_no_interpolate).all()

        # False in interpolate.ini

        mass_profile = ag.mp.SphericalIsothermal(einstein_radius=1.0)

        deflections_interpolate = mass_profile.deflections_from_grid(
            grid=grid_interpolate
        )

        grid_interp = mass_profile.deflections_from_grid(
            grid=grid_interpolate.grid_interp
        )
        interpolated_grid = grid_interpolate.interpolated_grid_from_grid_interp(
            grid_interp=grid_interp
        )
        assert (deflections_interpolate == interpolated_grid).all()



import autofit as af
from autogalaxy.pipeline.phase import abstract

from .hyper_phase import HyperPhase


# noinspection PyAbstractClass
class ModelFixingHyperPhase(HyperPhase):
    def __init__(
        self,
        phase: abstract.AbstractPhase,
        search,
        hyper_name: str,
        model_classes=tuple(),
    ):
        super().__init__(phase=phase, search=search, hyper_name=hyper_name)

        self.model_classes = model_classes

    def make_hyper_phase(self):
        phase = super().make_hyper_phase()

        return phase

    def make_model(self, instance):
        return instance.as_model(self.model_classes)

    def run_hyper(self, dataset, info=None, results=None, **kwargs):
        """
        Run the phase, overriding the search's model instance with one created to
        only fit pixelization hyperparameters.
        """

        self.results = results or af.ResultsCollection()

        phase = self.make_hyper_phase()
        phase.model = self.make_model(results.last.instance)

        return phase.run(dataset, mask=results.last.mask, results=results)


class InversionPhase(ModelFixingHyperPhase):
    def __init__(self, phase: abstract.AbstractPhase, search, model_classes=tuple()):
        super().__init__(
            phase=phase,
            search=search,
            model_classes=model_classes,
            hyper_name="inversion",
        )



import warnings

import autofit as af
import numpy as np
from autoarray.structures import arrays
from autoarray.structures import grids
from autogalaxy import dimensions as dim
from autogalaxy.profiles import mass_profiles as mp
from pyquad import quad_grid
from scipy.special import wofz
import typing


class StellarProfile:

    pass


class EllipticalGaussian(mp.EllipticalMassProfile, StellarProfile):
    @af.map_types
    def __init__(
        self,
        centre: dim.Position = (0.0, 0.0),
        elliptical_comps: typing.Tuple[float, float] = (0.0, 0.0),
        intensity: dim.Luminosity = 0.1,
        sigma: dim.Length = 0.01,
        mass_to_light_ratio: dim.MassOverLuminosity = 1.0,
    ):
        """ The elliptical Gaussian light profile.

        Parameters
        ----------
        centre : (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        elliptical_comps : (float, float)
            The first and second ellipticity components of the elliptical coordinate system, where
            fac = (1 - axis_ratio) / (1 + axis_ratio), ellip_y = fac * sin(2*phi) and ellip_x = fac * cos(2*phi).
        intensity : float
            Overall intensity normalisation of the light profiles (electrons per second).
        sigma : float
            The sigma value of the Gaussian.
        """

        super(EllipticalGaussian, self).__init__(
            centre=centre, elliptical_comps=elliptical_comps
        )
        super(mp.EllipticalMassProfile, self).__init__(
            centre=centre, elliptical_comps=elliptical_comps
        )
        self.mass_to_light_ratio = mass_to_light_ratio
        self.intensity = intensity
        self.sigma = sigma

        if self.axis_ratio > 0.9999:
            self.axis_ratio = 0.9999

    def omega_from_grid_and_q(self, grid_complex, q):
        x = np.real(grid_complex)
        y = np.imag(grid_complex)
        faddeeva = wofz((q * x) + (1j * y / q))
        return (
            np.exp(-x ** 2.0 - 2.0 * 1j * x * y) * np.exp(y ** 2.0)
            - (
                1j
                * np.exp(
                    -x ** 2.0 * (1.0 - q ** 2.0) - y ** 2.0 * ((1.0 / q ** 2.0) - 1.0)
                )
            )
            * faddeeva
        )

    def sigma_from_grid(self, grid):
        input_grid = (self.axis_ratio * (grid[:, 1] + 1j * grid[:, 0])) / (
            self.sigma * np.sqrt(2 * (1.0 - self.axis_ratio ** 2.0))
        )
        return self.omega_from_grid_and_q(
            grid_complex=input_grid, q=1
        ) - self.omega_from_grid_and_q(grid_complex=input_grid, q=self.axis_ratio)

    def deflections_from_grid(self, grid):

        with warnings.catch_warnings():
            warnings.filterwarnings("error")
            try:
                return self.deflections_from_grid_via_analytic(grid=grid)
            except RuntimeWarning:
                return self.deflections_from_grid_via_integrator(grid=grid)

    @grids.grid_like_to_structure
    @grids.transform
    @grids.relocate_to_radial_minimum
    def deflections_from_grid_via_analytic(self, grid):
        """
        Calculate the deflection angles at a given set of arc-second gridded coordinates.

        Parameters
        ----------
        grid : aa.Grid
            The grid of (y,x) arc-second coordinates the deflection angles are computed on.

        """

        deflections = (
            self.mass_to_light_ratio
            * self.intensity
            * (1.0 / self.sigma * np.sqrt(2.0 * np.pi))
            * self.sigma
            * np.sqrt((2 * np.pi) / (1.0 - self.axis_ratio ** 2.0))
            * self.sigma_from_grid(grid=grid)
        )

        return self.rotate_grid_from_profile(
            np.multiply(
                1.0, np.vstack((-1.0 * np.imag(deflections), np.real(deflections))).T
            )
        )

    @grids.grid_like_to_structure
    @grids.transform
    @grids.relocate_to_radial_minimum
    def deflections_from_grid_via_integrator(self, grid):
        """
        Calculate the deflection angles at a given set of arc-second gridded coordinates.

        Parameters
        ----------
        grid : aa.Grid
            The grid of (y,x) arc-second coordinates the deflection angles are computed on.

        """

        def calculate_deflection_component(npow, index):

            deflection_grid = self.axis_ratio * grid[:, index]
            deflection_grid *= (
                (1.0 / self.sigma * np.sqrt(2.0 * np.pi))
                * self.intensity
                * self.mass_to_light_ratio
                * quad_grid(
                    self.deflection_func,
                    0.0,
                    1.0,
                    grid,
                    args=(npow, self.axis_ratio, self.sigma),
                )[0]
            )

            return deflection_grid

        deflection_y = calculate_deflection_component(1.0, 0)
        deflection_x = calculate_deflection_component(0.0, 1)

        return self.rotate_grid_from_profile(
            np.multiply(1.0, np.vstack((deflection_y, deflection_x)).T)
        )

    @staticmethod
    def deflection_func(u, y, x, npow, axis_ratio, sigma):
        eta_u = np.sqrt(axis_ratio) * np.sqrt(
            (u * ((x ** 2) + (y ** 2 / (1 - (1 - axis_ratio ** 2) * u))))
        )

        return np.exp(-0.5 * np.square(np.divide(eta_u, sigma))) / (
            (1 - (1 - axis_ratio ** 2) * u) ** (npow + 0.5)
        )

    @grids.grid_like_to_structure
    @grids.transform
    @grids.relocate_to_radial_minimum
    def convergence_from_grid(self, grid):
        """ Calculate the projected convergence at a given set of arc-second gridded coordinates.

        Parameters
        ----------
        grid : aa.Grid
            The grid of (y,x) arc-second coordinates the convergence is computed on.

        """
        return self.convergence_func(self.grid_to_eccentric_radii(grid))

    def convergence_func(self, grid_radius):
        return self.mass_to_light_ratio * self.intensity_at_radius(grid_radius)

    def intensity_at_radius(self, grid_radii):
        """Calculate the intensity of the Gaussian light profile on a grid of radial coordinates.

        Parameters
        ----------
        grid_radii : float
            The radial distance from the centre of the profile. for each coordinate on the grid.
        """
        return np.multiply(
            np.divide(self.intensity, self.sigma * np.sqrt(2.0 * np.pi)),
            np.exp(-0.5 * np.square(np.divide(grid_radii, self.sigma))),
        )


# noinspection PyAbstractClass
class AbstractEllipticalSersic(mp.EllipticalMassProfile, StellarProfile):
    @af.map_types
    def __init__(
        self,
        centre: dim.Position = (0.0, 0.0),
        elliptical_comps: typing.Tuple[float, float] = (0.0, 0.0),
        intensity: dim.Luminosity = 0.1,
        effective_radius: dim.Length = 0.6,
        sersic_index: float = 0.6,
        mass_to_light_ratio: dim.MassOverLuminosity = 1.0,
    ):
        """
        The Sersic mass profile, the mass profiles of the light profiles that are used to fit and subtract the lens \
        model_galaxy's light.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        elliptical_comps : (float, float)
            The first and second ellipticity components of the elliptical coordinate system, where
            fac = (1 - axis_ratio) / (1 + axis_ratio), ellip_y = fac * sin(2*phi) and ellip_x = fac * cos(2*phi).
        intensity : float
            Overall flux intensity normalisation in the light profiles (electrons per second).
        effective_radius : float
            The radius containing half the light of this profile.
        sersic_index : float
            Controls the concentration of the of the profile (lower value -> less concentrated, \
            higher value -> more concentrated).
        mass_to_light_ratio : float
            The mass-to-light ratio of the light profiles
        """
        super(AbstractEllipticalSersic, self).__init__(
            centre=centre, elliptical_comps=elliptical_comps
        )
        super(mp.EllipticalMassProfile, self).__init__(
            centre=centre, elliptical_comps=elliptical_comps
        )
        self.mass_to_light_ratio = mass_to_light_ratio
        self.intensity = intensity
        self.effective_radius = effective_radius
        self.sersic_index = sersic_index

    @grids.grid_like_to_structure
    @grids.transform
    @grids.relocate_to_radial_minimum
    def convergence_from_grid(self, grid):
        """ Calculate the projected convergence at a given set of arc-second gridded coordinates.

        Parameters
        ----------
        grid : aa.Grid
            The grid of (y,x) arc-second coordinates the convergence is computed on.

        """
        return self.convergence_func(self.grid_to_eccentric_radii(grid))

    def convergence_func(self, grid_radius):
        return self.mass_to_light_ratio * self.intensity_at_radius(grid_radius)

    @grids.grid_like_to_structure
    def potential_from_grid(self, grid):
        return arrays.Array.manual_1d(
            array=np.zeros(shape=grid.shape[0]), shape_2d=grid.sub_shape_2d
        )

    @property
    def ellipticity_rescale(self):
        return 1.0 - ((1.0 - self.axis_ratio) / 2.0)

    def intensity_at_radius(self, radius):
        """ Compute the intensity of the profile at a given radius.

        Parameters
        ----------
        radius : float
            The distance from the centre of the profile.
        """
        return self.intensity * np.exp(
            -self.sersic_constant
            * (((radius / self.effective_radius) ** (1.0 / self.sersic_index)) - 1)
        )

    @property
    def sersic_constant(self):
        """ A parameter derived from Sersic index which ensures that effective radius contains 50% of the profile's
        total integrated light.
        """
        return (
            (2 * self.sersic_index)
            - (1.0 / 3.0)
            + (4.0 / (405.0 * self.sersic_index))
            + (46.0 / (25515.0 * self.sersic_index ** 2))
            + (131.0 / (1148175.0 * self.sersic_index ** 3))
            - (2194697.0 / (30690717750.0 * self.sersic_index ** 4))
        )

    @property
    def elliptical_effective_radius(self):
        """The effective_radius of a Sersic light profile is defined as the circular effective radius. This is the \
        radius within which a circular aperture contains half the profiles's total integrated light. For elliptical \
        systems, this won't robustly capture the light profile's elliptical shape.

        The elliptical effective radius instead describes the major-axis radius of the ellipse containing \
        half the light, and may be more appropriate for highly flattened systems like disk galaxies."""
        return self.effective_radius / np.sqrt(self.axis_ratio)

    @property
    def unit_mass(self):
        return self.mass_to_light_ratio.unit_mass


class EllipticalSersic(AbstractEllipticalSersic):
    @staticmethod
    def deflection_func(
        u, y, x, npow, axis_ratio, sersic_index, effective_radius, sersic_constant
    ):
        eta_u = np.sqrt(axis_ratio) * np.sqrt(
            (u * ((x ** 2) + (y ** 2 / (1 - (1 - axis_ratio ** 2) * u))))
        )

        return np.exp(
            -sersic_constant
            * (((eta_u / effective_radius) ** (1.0 / sersic_index)) - 1)
        ) / ((1 - (1 - axis_ratio ** 2) * u) ** (npow + 0.5))

    @grids.grid_like_to_structure
    @grids.transform
    @grids.relocate_to_radial_minimum
    def deflections_from_grid(self, grid):
        """
        Calculate the deflection angles at a given set of arc-second gridded coordinates.

        Parameters
        ----------
        grid : aa.Grid
            The grid of (y,x) arc-second coordinates the deflection angles are computed on.

        """

        def calculate_deflection_component(npow, index):
            sersic_constant = self.sersic_constant

            deflection_grid = self.axis_ratio * grid[:, index]
            deflection_grid *= (
                self.intensity
                * self.mass_to_light_ratio
                * quad_grid(
                    self.deflection_func,
                    0.0,
                    1.0,
                    grid,
                    args=(
                        npow,
                        self.axis_ratio,
                        self.sersic_index,
                        self.effective_radius,
                        sersic_constant,
                    ),
                )[0]
            )

            return deflection_grid

        deflection_y = calculate_deflection_component(1.0, 0)
        deflection_x = calculate_deflection_component(0.0, 1)

        return self.rotate_grid_from_profile(
            np.multiply(1.0, np.vstack((deflection_y, deflection_x)).T)
        )


class SphericalSersic(EllipticalSersic):
    @af.map_types
    def __init__(
        self,
        centre: dim.Position = (0.0, 0.0),
        intensity: dim.Luminosity = 0.1,
        effective_radius: dim.Length = 0.6,
        sersic_index: float = 0.6,
        mass_to_light_ratio: dim.MassOverLuminosity = 1.0,
    ):
        """
        The Sersic mass profile, the mass profiles of the light profiles that are used to fit and subtract the lens
        model_galaxy's light.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) arc-second coordinates of the profile centre
        intensity : float
            Overall flux intensity normalisation in the light profiles (electrons per second)
        effective_radius : float
            The circular radius containing half the light of this profile.
        sersic_index : float
            Controls the concentration of the of the profile (lower value -> less concentrated, \
            higher value -> more concentrated).
        mass_to_light_ratio : float
            The mass-to-light ratio of the light profile.
        """
        super(SphericalSersic, self).__init__(
            centre=centre,
            elliptical_comps=(0.0, 0.0),
            intensity=intensity,
            effective_radius=effective_radius,
            sersic_index=sersic_index,
            mass_to_light_ratio=mass_to_light_ratio,
        )


class EllipticalExponential(EllipticalSersic):
    @af.map_types
    def __init__(
        self,
        centre: dim.Position = (0.0, 0.0),
        elliptical_comps: typing.Tuple[float, float] = (0.0, 0.0),
        intensity: dim.Luminosity = 0.1,
        effective_radius: dim.Length = 0.6,
        mass_to_light_ratio: dim.MassOverLuminosity = 1.0,
    ):
        """
        The EllipticalExponential mass profile, the mass profiles of the light profiles that are used to fit and
        subtract the lens model_galaxy's light.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        elliptical_comps : (float, float)
            The first and second ellipticity components of the elliptical coordinate system, where
            fac = (1 - axis_ratio) / (1 + axis_ratio), ellip_y = fac * sin(2*phi) and ellip_x = fac * cos(2*phi).
        intensity : float
            Overall flux intensity normalisation in the light profiles (electrons per second).
        effective_radius : float
            The circular radius containing half the light of this profile.
        mass_to_light_ratio : float
            The mass-to-light ratio of the light profiles
        """
        super(EllipticalExponential, self).__init__(
            centre=centre,
            elliptical_comps=elliptical_comps,
            intensity=intensity,
            effective_radius=effective_radius,
            sersic_index=1.0,
            mass_to_light_ratio=mass_to_light_ratio,
        )


class SphericalExponential(EllipticalExponential):
    @af.map_types
    def __init__(
        self,
        centre: dim.Position = (0.0, 0.0),
        intensity: dim.Luminosity = 0.1,
        effective_radius: dim.Length = 0.6,
        mass_to_light_ratio: dim.MassOverLuminosity = 1.0,
    ):
        """
        The Exponential mass profile, the mass profiles of the light profiles that are used to fit and subtract the lens
        model_galaxy's light.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        intensity : float
            Overall flux intensity normalisation in the light profiles (electrons per second).
        effective_radius : float
            The circular radius containing half the light of this profile.
        mass_to_light_ratio : float
            The mass-to-light ratio of the light profiles.
        """
        super(SphericalExponential, self).__init__(
            centre=centre,
            elliptical_comps=(0.0, 0.0),
            intensity=intensity,
            effective_radius=effective_radius,
            mass_to_light_ratio=mass_to_light_ratio,
        )


class EllipticalDevVaucouleurs(EllipticalSersic):
    @af.map_types
    def __init__(
        self,
        centre: dim.Position = (0.0, 0.0),
        elliptical_comps: typing.Tuple[float, float] = (0.0, 0.0),
        intensity: dim.Luminosity = 0.1,
        effective_radius: dim.Length = 0.6,
        mass_to_light_ratio: dim.MassOverLuminosity = 1.0,
    ):
        """
        The EllipticalDevVaucouleurs mass profile, the mass profiles of the light profiles that are used to fit and
        subtract the lens model_galaxy's light.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        elliptical_comps : (float, float)
            The first and second ellipticity components of the elliptical coordinate system, where
            fac = (1 - axis_ratio) / (1 + axis_ratio), ellip_y = fac * sin(2*phi) and ellip_x = fac * cos(2*phi).
        intensity : float
            Overall flux intensity normalisation in the light profiles (electrons per second).
        effective_radius : float
            The radius containing half the light of this profile.
        mass_to_light_ratio : float
            The mass-to-light ratio of the light profile.
        """
        super(EllipticalDevVaucouleurs, self).__init__(
            centre=centre,
            elliptical_comps=elliptical_comps,
            intensity=intensity,
            effective_radius=effective_radius,
            sersic_index=4.0,
            mass_to_light_ratio=mass_to_light_ratio,
        )


class SphericalDevVaucouleurs(EllipticalDevVaucouleurs):
    @af.map_types
    def __init__(
        self,
        centre: dim.Position = (0.0, 0.0),
        intensity: dim.Luminosity = 0.1,
        effective_radius: dim.Length = 0.6,
        mass_to_light_ratio: dim.MassOverLuminosity = 1.0,
    ):
        """
        The DevVaucouleurs mass profile, the mass profiles of the light profiles that are used to fit and subtract the
        lens model_galaxy's light.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        intensity : float
            Overall flux intensity normalisation in the light profiles (electrons per second).
        effective_radius : float
            The circular radius containing half the light of this profile.
        mass_to_light_ratio : float
            The mass-to-light ratio of the light profiles.
        """
        super(SphericalDevVaucouleurs, self).__init__(
            centre=centre,
            elliptical_comps=(0.0, 0.0),
            intensity=intensity,
            effective_radius=effective_radius,
            mass_to_light_ratio=mass_to_light_ratio,
        )


class EllipticalSersicRadialGradient(AbstractEllipticalSersic):
    @af.map_types
    def __init__(
        self,
        centre: dim.Position = (0.0, 0.0),
        elliptical_comps: typing.Tuple[float, float] = (0.0, 0.0),
        intensity: dim.Luminosity = 0.1,
        effective_radius: dim.Length = 0.6,
        sersic_index: float = 0.6,
        mass_to_light_ratio: dim.MassOverLuminosity = 1.0,
        mass_to_light_gradient: float = 0.0,
    ):
        """
        Setup a Sersic mass and light profiles.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        elliptical_comps : (float, float)
            The first and second ellipticity components of the elliptical coordinate system, where
            fac = (1 - axis_ratio) / (1 + axis_ratio), ellip_y = fac * sin(2*phi) and ellip_x = fac * cos(2*phi).
        intensity : float
            Overall flux intensity normalisation in the light profiles (electrons per second).
        effective_radius : float
            The circular radius containing half the light of this profile.
        sersic_index : float
            Controls the concentration of the of the profile (lower value -> less concentrated, \
            higher value -> more concentrated).
        mass_to_light_ratio : float
            The mass-to-light ratio of the light profile.
        mass_to_light_gradient : float
            The mass-to-light radial gradient.
        """
        super(EllipticalSersicRadialGradient, self).__init__(
            centre=centre,
            elliptical_comps=elliptical_comps,
            intensity=intensity,
            effective_radius=effective_radius,
            sersic_index=sersic_index,
            mass_to_light_ratio=mass_to_light_ratio,
        )
        self.mass_to_light_gradient = mass_to_light_gradient

    @grids.grid_like_to_structure
    @grids.transform
    @grids.relocate_to_radial_minimum
    def convergence_from_grid(self, grid):
        """ Calculate the projected convergence at a given set of arc-second gridded coordinates.

        Parameters
        ----------
        grid : aa.Grid
            The grid of (y,x) arc-second coordinates the convergence is computed on.

        """
        return self.convergence_func(self.grid_to_eccentric_radii(grid))

    @grids.grid_like_to_structure
    @grids.transform
    @grids.relocate_to_radial_minimum
    def deflections_from_grid(self, grid):
        """
        Calculate the deflection angles at a given set of arc-second gridded coordinates.

        Parameters
        ----------
        grid : aa.Grid
            The grid of (y,x) arc-second coordinates the deflection angles are computed on.

        """

        def calculate_deflection_component(npow, index):
            sersic_constant = self.sersic_constant

            deflection_grid = self.axis_ratio * grid[:, index]
            deflection_grid *= (
                self.intensity
                * self.mass_to_light_ratio
                * quad_grid(
                    self.deflection_func,
                    0.0,
                    1.0,
                    grid,
                    args=(
                        npow,
                        self.axis_ratio,
                        self.sersic_index,
                        self.effective_radius,
                        self.mass_to_light_gradient,
                        sersic_constant,
                    ),
                )[0]
            )
            return deflection_grid

        deflection_y = calculate_deflection_component(1.0, 0)
        deflection_x = calculate_deflection_component(0.0, 1)

        return self.rotate_grid_from_profile(
            np.multiply(1.0, np.vstack((deflection_y, deflection_x)).T)
        )

    def convergence_func(self, grid_radius):
        return (
            self.mass_to_light_ratio
            * (
                ((self.axis_ratio * grid_radius) / self.effective_radius)
                ** -self.mass_to_light_gradient
            )
            * self.intensity_at_radius(grid_radius)
        )

    @staticmethod
    def deflection_func(
        u,
        y,
        x,
        npow,
        axis_ratio,
        sersic_index,
        effective_radius,
        mass_to_light_gradient,
        sersic_constant,
    ):
        eta_u = np.sqrt(axis_ratio) * np.sqrt(
            (u * ((x ** 2) + (y ** 2 / (1 - (1 - axis_ratio ** 2) * u))))
        )

        return (
            (((axis_ratio * eta_u) / effective_radius) ** -mass_to_light_gradient)
            * np.exp(
                -sersic_constant
                * (((eta_u / effective_radius) ** (1.0 / sersic_index)) - 1)
            )
            / ((1 - (1 - axis_ratio ** 2) * u) ** (npow + 0.5))
        )


class SphericalSersicRadialGradient(EllipticalSersicRadialGradient):
    @af.map_types
    def __init__(
        self,
        centre: dim.Position = (0.0, 0.0),
        intensity: dim.Luminosity = 0.1,
        effective_radius: dim.Length = 0.6,
        sersic_index: float = 0.6,
        mass_to_light_ratio: dim.MassOverLuminosity = 1.0,
        mass_to_light_gradient: float = 0.0,
    ):
        """
        Setup a Sersic mass and light profiles.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        intensity : float
            Overall flux intensity normalisation in the light profiles (electrons per second).
        effective_radius : float
            The circular radius containing half the light of this profile.
        sersic_index : float
            Controls the concentration of the of the profile (lower value -> less concentrated, \
            higher value -> more concentrated).
        mass_to_light_ratio : float
            The mass-to-light ratio of the light profile.
        mass_to_light_gradient : float
            The mass-to-light radial gradient.
        """
        super(SphericalSersicRadialGradient, self).__init__(
            centre=centre,
            elliptical_comps=(0.0, 0.0),
            intensity=intensity,
            effective_radius=effective_radius,
            sersic_index=sersic_index,
            mass_to_light_ratio=mass_to_light_ratio,
            mass_to_light_gradient=mass_to_light_gradient,
        )



import autogalaxy as ag
import numpy as np
import pytest

# Arc second coordinate grid is:.

# [[[-2.0, -2.0], [-2.0, -1.0], [-2.0, 0.0], [-2.0, 1.0], [-2.0, 2.0]],
# [[[-1.0, -2.0], [-1.0, -1.0], [-1.0, 0.0], [-1.0, 1.0], [-1.0, 2.0]],
# [[[ 0.0, -2.0], [ 0.0, -1.0], [ 0.0, 0.0], [ 0.0, 1.0], [ 0.0, 2.0]],
# [[[ 1.0, -2.0], [ 1.0, -1.0], [ 1.0, 0.0], [ 1.0, 1.0], [ 1.0, 2.0]],
# [[[ 2.0, -2.0], [ 2.0, -1.0], [ 2.0, 0.0], [ 2.0, 1.0], [ 2.0, 2.0]],


def test__centre_light_profile_on_grid_coordinate__peak_flux_is_correct_index():

    grid = ag.Grid.uniform(shape_2d=(5, 5), pixel_scales=1.0, sub_size=1)

    sersic = ag.lp.SphericalSersic(centre=(2.0, -2.0))
    image = sersic.image_from_grid(grid=grid)

    assert image.in_1d.argmax() == 0
    assert np.unravel_index(image.in_2d.argmax(), image.in_2d.shape) == (0, 0)

    sersic = ag.lp.SphericalSersic(centre=(2.0, 2.0))
    image = sersic.image_from_grid(grid=grid)

    assert image.in_1d.argmax() == 4
    assert np.unravel_index(image.in_2d.argmax(), image.in_2d.shape) == (0, 4)

    sersic = ag.lp.SphericalSersic(centre=(-2.0, -2.0))
    image = sersic.image_from_grid(grid=grid)

    assert image.in_1d.argmax() == 20
    assert np.unravel_index(image.in_2d.argmax(), image.in_2d.shape) == (4, 0)

    sersic = ag.lp.SphericalSersic(centre=(-2.0, 2.0))
    image = sersic.image_from_grid(grid=grid)

    assert image.in_1d.argmax() == 24
    assert np.unravel_index(image.in_2d.argmax(), image.in_2d.shape) == (4, 4)


def test__centre_mass_profile_on_grid_coordinate__peak_density_is_correct_index():

    grid = ag.Grid.uniform(shape_2d=(5, 5), pixel_scales=1.0, sub_size=1)

    sis = ag.mp.SphericalIsothermal(centre=(2.0, -2.0))
    density = sis.convergence_from_grid(grid=grid)

    assert density.in_1d.argmax() == 0
    assert np.unravel_index(density.in_2d.argmax(), density.in_2d.shape) == (0, 0)

    sis = ag.mp.SphericalIsothermal(centre=(2.0, 2.0))
    density = sis.convergence_from_grid(grid=grid)

    assert density.in_1d.argmax() == 4
    assert np.unravel_index(density.in_2d.argmax(), density.in_2d.shape) == (0, 4)

    sis = ag.mp.SphericalIsothermal(centre=(-2.0, -2.0))
    density = sis.convergence_from_grid(grid=grid)

    assert density.in_1d.argmax() == 20
    assert np.unravel_index(density.in_2d.argmax(), density.in_2d.shape) == (4, 0)

    sis = ag.mp.SphericalIsothermal(centre=(-2.0, 2.0))
    density = sis.convergence_from_grid(grid=grid)

    assert density.in_1d.argmax() == 24
    assert np.unravel_index(density.in_2d.argmax(), density.in_2d.shape) == (4, 4)


def test__deflection_angles():

    grid = ag.Grid.uniform(shape_2d=(5, 5), pixel_scales=1.0, sub_size=1)

    sis = ag.mp.SphericalIsothermal(centre=(0.0, 0.0), einstein_radius=1.0)
    deflections_y_2d = sis.deflections_from_grid(grid=grid).in_2d[:, :, 0]

    assert deflections_y_2d[0, 0] == pytest.approx(-1.0 * deflections_y_2d[4, 0], 1e-2)
    assert deflections_y_2d[1, 1] == pytest.approx(-1.0 * deflections_y_2d[3, 1], 1e-2)
    assert deflections_y_2d[1, 3] == pytest.approx(-1.0 * deflections_y_2d[3, 3], 1e-2)
    assert deflections_y_2d[0, 4] == pytest.approx(-1.0 * deflections_y_2d[4, 4], 1e-2)
    assert deflections_y_2d[2, 0] == pytest.approx(deflections_y_2d[2, 4], 1e-2)

    sis = ag.mp.SphericalIsothermal(centre=(0.0, 0.0), einstein_radius=1.0)
    deflections_x_2d = sis.deflections_from_grid(grid=grid).in_2d[:, :, 1]

    assert deflections_x_2d[0, 0] == pytest.approx(-1.0 * deflections_x_2d[0, 4], 1e-2)
    assert deflections_x_2d[1, 1] == pytest.approx(-1.0 * deflections_x_2d[1, 3], 1e-2)
    assert deflections_x_2d[3, 1] == pytest.approx(-1.0 * deflections_x_2d[3, 3], 1e-2)
    assert deflections_x_2d[4, 0] == pytest.approx(-1.0 * deflections_x_2d[4, 4], 1e-2)
    assert deflections_x_2d[0, 2] == pytest.approx(deflections_x_2d[4, 2], 1e-2)
